# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

from typing import Optional, Tuple, Union

from torch import Tensor
from torch.nn import Module

from brevitas.export.manager import _set_proxy_export_handler
from brevitas.export.manager import _set_proxy_export_mode
from brevitas.export.manager import _set_recurrent_layer_export_handler
from brevitas.export.manager import _set_recurrent_layer_export_mode
from brevitas.export.onnx.debug import DebugMarkerFunction
from brevitas.export.onnx.manager import ONNXBaseManager

from .function import BrevitasBinaryQuantFn
from .function import BrevitasQuantFn
from .function import BrevitasQuantLSTMCellFn
from .function import BrevitasTruncFn
from .handler import BrevitasActQuantProxyHandler
from .handler import BrevitasBiasQuantProxyHandler
from .handler import BrevitasDecoupledWeightQuantProxyHandler
from .handler import BrevitasDecoupledWeightQuantWithInputProxyHandler
from .handler import BrevitasQuantLSTMLayerHandler
from .handler import BrevitasTruncQuantProxyHandler
from .handler import BrevitasWeightQuantProxyHandler


class QONNXManager(ONNXBaseManager):
    target_name = 'brevitas'
    dequantize_tracing_input = False

    onnx_passes = [
        # use initializers instead of Constant nodes for fixed params
        "extract_constant_to_initializer",  # remove unused graph inputs & initializers
        "eliminate_unused_initializer"]

    handlers = [
        BrevitasActQuantProxyHandler,
        BrevitasBiasQuantProxyHandler,
        BrevitasWeightQuantProxyHandler,
        BrevitasDecoupledWeightQuantProxyHandler,
        BrevitasDecoupledWeightQuantWithInputProxyHandler,
        BrevitasTruncQuantProxyHandler,
        BrevitasQuantLSTMLayerHandler]

    custom_fns = [
        DebugMarkerFunction,
        BrevitasQuantFn,
        BrevitasBinaryQuantFn,
        BrevitasTruncFn,
        BrevitasQuantLSTMCellFn]

    @classmethod
    def set_export_mode(cls, model: Module, enabled: bool):
        _set_proxy_export_mode(model, enabled)
        _set_recurrent_layer_export_mode(model, enabled)

    @classmethod
    def set_export_handler(cls, module: Module):
        _set_proxy_export_handler(cls, module)
        _set_recurrent_layer_export_handler(cls, module)



import brevitas.nn as qnn
from brevitas.quant.experimental.float_quant_ocp import Fp8e5m2OCPWeight
from brevitas.quant.experimental.float_quant_ocp import Fp8e5m2OCPAct
from brevitas.quant.experimental.float_quant_ocp import Fp8e4m3OCPWeight
from brevitas.quant.experimental.float_quant_ocp import Fp8e4m3OCPAct

float8_conv = qnn.QuantConv2d(3, 8, weight_quant=Fp8e5m2OCPWeight, input_quant=Fp8e5m2OCPAct)


# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

from brevitas.core.function_wrapper import FloatClamp
from brevitas.core.function_wrapper import TensorClamp
from brevitas.core.quant.float import FloatQuant
from brevitas.core.scaling.float_scaling import FloatScaling
from brevitas.inject import ExtendedInjector
from brevitas.inject import value
from brevitas.proxy import ActFloatQuantProxyFromInjector
from brevitas.proxy import WeightFloatQuantProxyFromInjector
from brevitas.quant.solver import ActQuantSolver
from brevitas.quant.solver import WeightQuantSolver
from brevitas.quant.solver.common import SolveTensorQuantFloatToIntImplFromEnum


class FloatBase(SolveTensorQuantFloatToIntImplFromEnum):
    tensor_quant = FloatQuant
    signed = True
    float_to_int_impl_type = 'round'
    scaling_min_val = 1e-10
    float_clamp_impl = FloatClamp
    tensor_clamp_impl = TensorClamp

    @value
    def exponent_bias(exponent_bit_width):
        return 2 ** (exponent_bit_width - 1) - 1


class FloatWeightBase(FloatBase):
    proxy_class = WeightFloatQuantProxyFromInjector


class FloatActBase(FloatBase):
    proxy_class = ActFloatQuantProxyFromInjector


class ScaledFloatWeightBase(FloatWeightBase, WeightQuantSolver):
    scaling_stats_op = 'max'
    scaling_impl_type = 'stats'
    restrict_scaling_type = 'fp'
    float_scaling_impl = FloatScaling


class ScaledFloatActBase(FloatActBase, ActQuantSolver):
    scaling_stats_op = 'max'
    scaling_impl_type = 'parameter_from_stats'
    restrict_scaling_type = 'fp'
    collect_stats_steps = 300
    float_scaling_impl = FloatScaling


class Fp8e4m3Mixin(ExtendedInjector):
    bit_width = 8
    exponent_bit_width = 4
    mantissa_bit_width = 3
    saturating = True


class Fp8e5m2Mixin(ExtendedInjector):
    bit_width = 8
    exponent_bit_width = 5
    mantissa_bit_width = 2
    saturating = True


class Fp6e3m2Mixin(ExtendedInjector):
    bit_width = 6
    exponent_bit_width = 3
    mantissa_bit_width = 2
    saturating = True


class Fp6e2m3Mixin(ExtendedInjector):
    bit_width = 6
    exponent_bit_width = 2
    mantissa_bit_width = 3
    saturating = True


class Fp4e2m1Mixin(ExtendedInjector):
    bit_width = 4
    exponent_bit_width = 2
    mantissa_bit_width = 1
    saturating = True



from typing import Any, Optional, Tuple, Union

import torch

from brevitas.proxy.runtime_quant import ActQuantProxyFromInjector
from brevitas.quant_tensor import GroupwiseIntQuantTensor
from brevitas.utils.quant_utils import _CachedIOGroupwiseInt


class GroupwiseActQuantProxyFromInjector(ActQuantProxyFromInjector):

    def __init__(self, quant_layer, quant_injector):
        super().__init__(quant_layer, quant_injector)
        self.cache_class = _CachedIOGroupwiseInt

    @property
    def group_dim(self):
        return self.quant_injector.group_dim

    @property
    def group_size(self):
        return self.quant_injector.group_size

    def apply_input_view(self, x):
        x = super().apply_input_view(x)
        start_dim = self.group_dim if self.group_dim != -1 else -2
        return x.flatten(start_dim, start_dim + 1)

    def create_quant_tensor(
            self,
            qt_args: Union[torch.Tensor, Tuple[Any]],
            x: Optional[GroupwiseIntQuantTensor] = None) -> GroupwiseIntQuantTensor:
        if x is None:
            value, scale, zero_point, bit_width = qt_args
            out = GroupwiseIntQuantTensor(
                value,
                scale,
                zero_point,
                self.group_size,
                self.group_dim,
                bit_width,
                self.is_signed,
                self.training)
        else:
            out = GroupwiseIntQuantTensor(
                qt_args,
                x.scale,
                x.zero_point,
                self.group_size,
                self.group_dim,
                x.bit_width,
                x.signed,
                self.training)
        return out



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

from typing import Optional

import torch
from torch import Tensor
from torch.nn import Embedding
from torch.nn.functional import embedding

from brevitas.inject.defaults import Int8WeightPerTensorFloat
from brevitas.quant_tensor import _unpack_quant_tensor
from brevitas.quant_tensor import QuantTensor

from .mixin.parameter import QuantWeightMixin
from .quant_layer import WeightQuantType

__all__ = ['QuantEmbedding']


class QuantEmbedding(QuantWeightMixin, Embedding):

    def __init__(
            self,
            num_embeddings: int,
            embedding_dim: int,
            padding_idx: Optional[int] = None,
            max_norm: Optional[float] = None,
            norm_type: float = 2.,
            scale_grad_by_freq: bool = False,
            sparse: bool = False,
            _weight: Optional[Tensor] = None,
            weight_quant: WeightQuantType = Int8WeightPerTensorFloat,
            return_quant_tensor=False,
            device: Optional[torch.device] = None,
            dtype: Optional[torch.dtype] = None,
            **kwargs) -> None:
        Embedding.__init__(
            self,
            num_embeddings=num_embeddings,
            embedding_dim=embedding_dim,
            padding_idx=padding_idx,
            max_norm=max_norm,
            norm_type=norm_type,
            scale_grad_by_freq=scale_grad_by_freq,
            sparse=sparse,
            _weight=_weight,
            device=device,
            dtype=dtype)
        QuantWeightMixin.__init__(self, weight_quant=weight_quant, **kwargs)
        self.accept_quant_tensor = False
        self.return_quant_tensor = return_quant_tensor

    @property
    def output_channel_dim(self) -> int:
        return 0

    @property
    def out_channels(self) -> int:
        return self.num_embeddings

    def forward(self, inp):
        quant_weight = self.quant_weight()
        out = embedding(
            inp,
            quant_weight,
            self.padding_idx,
            self.max_norm,
            self.norm_type,
            self.scale_grad_by_freq,
            self.sparse)
        if self.return_quant_tensor:
            assert isinstance(out, QuantTensor), "Enable weight quantization to return QuantTensor"
            return out
        else:
            out = _unpack_quant_tensor(out)

        return out



"""
Adapted from https://github.com/IST-DASLab/gptq, released under the following LICENSE:

Copyright 2023 IST-DASLab

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import random

from datasets import load_dataset
import torch
from transformers import AutoTokenizer


def get_c4(nsamples, seed, seqlen, tokenizer, split='train', nvalsamples=0):
    if split == 'train':
        data = load_dataset(
            'allenai/c4',
            'allenai--c4',
            data_files={'train': 'en/c4-train.00000-of-01024.json.gz'},
            split='train',
            use_auth_token=False)

        random.seed(seed)
        dataloader = []
        for _ in range(nsamples):
            while True:
                i = random.randint(0, len(data) - 1)
                trainenc = tokenizer(data[i]['text'], return_tensors='pt')
                if trainenc.input_ids.shape[1] >= seqlen:
                    break
            i = random.randint(0, trainenc.input_ids.shape[1] - seqlen - 1)
            j = i + seqlen
            inp = trainenc.input_ids[:, i:j]
            dataloader.append(inp)
        return dataloader
    elif split == 'validation':
        data = load_dataset(
            'allenai/c4',
            'allenai--c4',
            data_files={'validation': 'en/c4-validation.00000-of-00008.json.gz'},
            split='validation',
            use_auth_token=False)

        random.seed(0)  # hardcoded for validation reproducibility
        valenc = []
        for _ in range(nvalsamples):
            while True:
                i = random.randint(0, len(data) - 1)
                tmp = tokenizer(data[i]['text'], return_tensors='pt')
                if tmp.input_ids.shape[1] >= seqlen:
                    break
            i = random.randint(0, tmp.input_ids.shape[1] - seqlen - 1)
            j = i + seqlen
            valenc.append(tmp.input_ids[:, i:j])

        valenc = torch.hstack(valenc)
        return valenc


def get_wikitext2(nsamples, seed, seqlen, tokenizer, type='raw', split='train'):
    from datasets import load_dataset
    dataset_name = 'wikitext-2-v1'
    if type == 'raw':
        dataset_name = 'wikitext-2-raw-v1'
    if split == 'train':
        traindata = load_dataset('wikitext', dataset_name, split='train')

        trainenc = tokenizer("\n\n".join(traindata['text']), return_tensors='pt')

        import random
        random.seed(seed)
        trainloader = []
        for _ in range(nsamples):
            i = random.randint(0, trainenc.input_ids.shape[1] - seqlen - 1)
            j = i + seqlen
            inp = trainenc.input_ids[:, i:j]
            attention_mask = torch.ones_like(inp)
            trainloader.append({'input_ids': inp, 'attention_mask': attention_mask})
        return trainloader
    elif split == 'validation':
        testdata = load_dataset('wikitext', dataset_name, split='test')
        testenc = tokenizer("\n\n".join(testdata['text']), return_tensors='pt')
        return testenc



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

import argparse
import copy
from hashlib import sha256
import json
import os
import pprint
import random

import numpy as np
import torch
import torch.nn as nn
import torch.optim as optim
import torch.optim.lr_scheduler as lrs

from brevitas_examples.super_resolution.models import get_model_by_name
from brevitas_examples.super_resolution.utils import device
from brevitas_examples.super_resolution.utils import evaluate_accumulator_bit_widths
from brevitas_examples.super_resolution.utils import evaluate_avg_psnr
from brevitas_examples.super_resolution.utils import export
from brevitas_examples.super_resolution.utils import get_bsd300_dataloaders
from brevitas_examples.super_resolution.utils import train_for_epoch

random_seed = 123456

random.seed(random_seed)
np.random.seed(random_seed)
torch.manual_seed(random_seed)

desc = """Training single-image super resolution models on the BSD300 dataset.

Example:
>> python train_model.py --data_root=data/ --model=quant_espcn_x2_w8a8_a2q_16b
"""

parser = argparse.ArgumentParser(description='PyTorch BSD300 Validation')
parser.add_argument('--data_root', help='Path to folder containing BSD300 val folder')
parser.add_argument(
    '--save_path', type=str, default='outputs/', help='Save path for exported model')
parser.add_argument(
    '--model',
    type=str,
    default='quant_espcn_x2_w8a8_a2q_16b',
    help='Name of the model configuration')
parser.add_argument('--workers', type=int, default=0, help='Number of data loading workers')
parser.add_argument('--batch_size', type=int, default=8, help='Minibatch size')
parser.add_argument('--learning_rate', type=float, default=1e-3, help='Learning rate')
parser.add_argument('--total_epochs', type=int, default=500, help='Total number of training epochs')
parser.add_argument('--weight_decay', type=float, default=1e-5, help='Weight decay')
parser.add_argument('--step_size', type=int, default=1)
parser.add_argument('--gamma', type=float, default=0.999)
parser.add_argument('--eval_acc_bw', action='store_true', default=False)
parser.add_argument('--save_pth_ckpt', action='store_true', default=False)
parser.add_argument('--save_model_io', action='store_true', default=False)
parser.add_argument('--export_to_qonnx', action='store_true', default=False)
parser.add_argument('--export_to_qcdq_onnx', action='store_true', default=False)
parser.add_argument('--export_to_qcdq_torch', action='store_true', default=False)


def filter_params(named_params, decay):
    decay_params, no_decay_params = [], []
    for name, param in named_params:
        # Do not apply weight decay to the bias or any scaling parameters
        if 'scaling' in name or name.endswith(".bias"):
            no_decay_params.append(param)
        else:
            decay_params.append(param)
    return [{
        'params': no_decay_params, 'weight_decay': 0.}, {
            'params': decay_params, 'weight_decay': decay}]


def main():
    args = parser.parse_args()

    # initialize model, dataset, and training environment
    model = get_model_by_name(args.model)
    model = model.to(device)
    trainloader, testloader = get_bsd300_dataloaders(
        args.data_root,
        num_workers=args.workers,
        batch_size=args.batch_size,
        upscale_factor=model.upscale_factor,
        download=True)
    criterion = nn.MSELoss()
    optimizer = optim.Adam(
        filter_params(model.named_parameters(), args.weight_decay),
        lr=args.learning_rate,
        weight_decay=args.weight_decay)
    scheduler = lrs.StepLR(optimizer, step_size=args.step_size, gamma=args.gamma)

    # train model
    best_psnr, best_weights = 0., copy.deepcopy(model.state_dict())
    for ep in range(args.total_epochs):
        train_loss = train_for_epoch(trainloader, model, criterion, optimizer)
        test_psnr = evaluate_avg_psnr(testloader, model)
        scheduler.step()
        print(f"[Epoch {ep:03d}] train_loss={train_loss:.4f}, test_psnr={test_psnr:.2f}")
        if test_psnr >= best_psnr:
            best_weights = copy.deepcopy(model.state_dict())
            best_psnr = test_psnr
    model.load_state_dict(best_weights)
    model = model.to(device)
    test_psnr = evaluate_avg_psnr(testloader, model)
    print(f"Final test_psnr={test_psnr:.2f}")

    # save checkpoint
    os.makedirs(args.save_path, exist_ok=True)
    if args.save_pth_ckpt:
        ckpt_path = f"{args.save_path}/{args.model}.pth"
        torch.save(best_weights, ckpt_path)
        with open(ckpt_path, "rb") as _file:
            bytes = _file.read()
            model_tag = sha256(bytes).hexdigest()[:8]
        new_ckpt_path = f"{args.save_path}/{args.model}-{model_tag}.pth"
        os.rename(ckpt_path, new_ckpt_path)
        print(f"Saved model checkpoint to {new_ckpt_path}")

    # evaluate accumulator bit widths
    if args.eval_acc_bw:
        inp = testloader.dataset[0][0].unsqueeze(0).to(device)
        stats_dict = {
            'acc_bit_widths': evaluate_accumulator_bit_widths(model, inp),
            'performance': {
                'test_psnr': test_psnr, 'train_loss': train_loss}}
        with open(f"{args.save_path}/stats.json", "w") as outfile:
            json.dump(stats_dict, outfile, indent=4)
        pretty_stats_dict = pprint.pformat(stats_dict, sort_dicts=False)
        print(pretty_stats_dict)

    # save and export model
    export(model, testloader, args)


if __name__ == '__main__':
    main()



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

from abc import ABC
from abc import abstractmethod
import math
from warnings import warn

import torch
from torch import Tensor
from torch.nn import Module

from brevitas.function.ops import max_int
from brevitas.function.ops import min_int

__all__ = [
    'BaseHandler', 'BitWidthHandlerMixin', 'ZeroPointHandlerMixin', 'FloatZeroPointHandlerMixin']


class BaseHandler(Module, ABC):

    def __init__(self) -> None:
        super().__init__()

    def attach_debug_info(self, module):
        pass

    @abstractmethod
    def prepare_for_export(self, module):
        pass


class QuantAxisMixin(ABC):

    @classmethod
    def quant_axis(cls, scale):
        for i, s in enumerate(scale.shape):
            if s != 1:
                return i
        return None


class FloatClipMixin(ABC):

    @classmethod
    def clip_symbolic_kwargs(cls, narrow, signed, exponent_bit_width, mantissa_bit_width):
        return None


class ClipMixin(ABC):

    @classmethod
    def int_clip_symbolic_kwargs(cls, narrow, signed, bit_width):
        # equality comparisons among power-of-2 floats are okay
        if narrow or bit_width != 8. and bit_width != 32.:
            if signed and (bit_width < 8. or narrow and bit_width <= 8.):
                dtype = torch.int8
            elif not signed and (bit_width < 8. or narrow and bit_width <= 8.):
                dtype = torch.uint8
            elif signed and (bit_width < 32. or narrow and bit_width <= 32.):
                dtype = torch.int32
            else:
                raise RuntimeError(
                    f"Sign {signed} and bit width {bit_width} not supported for export.")
            return {
                'min_val': min_int(signed, narrow, bit_width).to(dtype),
                'max_val': max_int(signed, narrow, bit_width).to(dtype)}
        else:
            return None


class BitWidthHandlerMixin(ABC):

    @classmethod
    def validate_bit_width(cls, bit_width: Tensor, reference: int, le_then=False):
        if bit_width is None:
            raise RuntimeError("Bit width cannot be None")
        if isinstance(bit_width, torch.Tensor):
            bit_width = bit_width.item()
        bit_width = int(bit_width)
        if bit_width > reference:
            raise RuntimeError(f"Bit width {bit_width} is not supported.")
        elif bit_width < reference and not le_then:
            raise RuntimeError(f"Bit width {bit_width} is not supported, should be {reference}b.")
        return bit_width

    @classmethod
    def validate_8b_bit_width(cls, bit_width: Tensor, le_then=False):
        return cls.validate_bit_width(bit_width, 8, le_then)

    @classmethod
    def validate_16b_bit_width(cls, bit_width: Tensor, le_then=False):
        return cls.validate_bit_width(bit_width, 16, le_then)

    @classmethod
    def validate_32b_bit_width(cls, bit_width: Tensor, le_then=False):
        return cls.validate_bit_width(bit_width, 32, le_then)


class ScaleHandlerMixin(ABC):

    @classmethod
    def validate_scalar_scale(cls, scale: Tensor):
        if scale is None:
            raise RuntimeError("Scale cannot be None.")
        if scale.view(-1).shape[0] != 1:
            raise RuntimeError("Only per-tensor scaling is supported.")
        return scale.item()

    @classmethod
    def validate_scalar_int_exponent(cls, scale: Tensor):
        cls.validate_scalar_scale(scale)
        exponent = math.log2(scale)
        if not exponent.is_integer():
            raise RuntimeError("Only power-of-two scale factors are supported.")
        exponent = int(exponent)
        return exponent

    @classmethod
    def validate_neg_scalar_int_exponent(cls, scale: Tensor):
        return -cls.validate_scalar_int_exponent(scale)


class FloatZeroPointHandlerMixin(ABC):

    @classmethod
    def zero_point_with_dtype(
            cls, exponent_bit_width, mantissa_bit_width, is_ocp, is_fnuz, zero_point):
        if is_ocp:
            if exponent_bit_width == 4 and mantissa_bit_width == 3:
                return zero_point.type(torch.float8_e4m3fn)
            elif exponent_bit_width == 5 and mantissa_bit_width == 2:
                return zero_point.type(torch.float8_e5m2)
        elif is_fnuz:
            if exponent_bit_width == 4 and mantissa_bit_width == 3:
                return zero_point.type(torch.float8_e4m3fnuz)
            elif exponent_bit_width == 5 and mantissa_bit_width == 2:
                return zero_point.type(torch.float8_e5m2fnuz)
        else:
            raise NotImplementedError


class ZeroPointHandlerMixin(ABC):

    @classmethod
    def zero_point_with_dtype(cls, signed, bit_width, zero_point):
        if not signed:
            if (zero_point < 0).any():
                raise RuntimeError("Zero points have to be positive under unsigned quantization")
            if bit_width > 8:
                raise RuntimeError("Unsigned zero-point with bit-width > 8 not supported.")
            return zero_point.type(torch.uint8)
        else:
            if bit_width <= 8:
                return zero_point.type(torch.int8)
            else:
                return zero_point.type(torch.int32)



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

import math
from typing import Optional, Tuple, Type, Union

import torch
from torch import Tensor
from torch.nn import Conv1d
from torch.nn import Conv2d
from torch.nn import Conv3d
from torch.nn import functional as F

from brevitas.function.ops import max_int
from brevitas.function.ops_ste import ceil_ste
from brevitas.inject.defaults import Int8WeightPerTensorFloat
from brevitas.quant_tensor import QuantTensor

from .quant_layer import ActQuantType
from .quant_layer import BiasQuantType
from .quant_layer import QuantWeightBiasInputOutputLayer as QuantWBIOL
from .quant_layer import WeightQuantType

__all__ = ['QuantConv1d', 'QuantConv2d', 'QuantConv3d']


class QuantConv1d(QuantWBIOL, Conv1d):

    def __init__(
            self,
            in_channels: int,
            out_channels: int,
            kernel_size: Union[int, Tuple[int, int]],
            stride: Union[int, Tuple[int, int]] = 1,
            padding: Union[int, Tuple[int, int]] = 0,
            dilation: Union[int, Tuple[int, int]] = 1,
            groups: int = 1,
            padding_mode: str = 'zeros',
            bias: Optional[bool] = True,
            weight_quant: Optional[WeightQuantType] = Int8WeightPerTensorFloat,
            bias_quant: Optional[BiasQuantType] = None,
            input_quant: Optional[ActQuantType] = None,
            output_quant: Optional[ActQuantType] = None,
            return_quant_tensor: bool = False,
            device: Optional[torch.device] = None,
            dtype: Optional[torch.dtype] = None,
            **kwargs) -> None:
        # avoid an init error in the super class by setting padding to 0
        if padding_mode == 'zeros' and padding == 'same' and (stride > 1 if isinstance(
                stride, int) else any(map(lambda x: x > 1, stride))):
            padding = 0
            is_same_padded_strided = True
        else:
            is_same_padded_strided = False
        Conv1d.__init__(
            self,
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            dilation=dilation,
            groups=groups,
            padding_mode=padding_mode,
            bias=bias,
            device=device,
            dtype=dtype)
        QuantWBIOL.__init__(
            self,
            weight_quant=weight_quant,
            bias_quant=bias_quant,
            input_quant=input_quant,
            output_quant=output_quant,
            return_quant_tensor=return_quant_tensor,
            **kwargs)
        self.is_same_padded_strided = is_same_padded_strided

    @property
    def per_elem_ops(self):
        return 2 * self.kernel_size[0] * (self.in_channels // self.groups)

    @property
    def output_channel_dim(self):
        if self.transposed:
            return 1
        else:
            return 0

    @property
    def channelwise_separable(self) -> bool:
        return self.groups == self.in_channels

    def conv1d_same_zeros_pad_stride(self, x, weight, bias):
        ih = x.size()[-1]
        kh = weight.size()[-1]
        sh = self.stride[0]
        oh = math.ceil(ih / sh)
        pad_h = max((oh - 1) * self.stride[0] + (kh - 1) * self.dilation[0] + 1 - ih, 0)
        if pad_h > 0:
            x = F.pad(x, [pad_h // 2, pad_h - pad_h // 2])
        out = F.conv1d(x, weight, bias, self.stride, 0, self.dilation, self.groups)
        return out

    def forward(self, input: Union[Tensor, QuantTensor]) -> Union[Tensor, QuantTensor]:
        return self.forward_impl(input)

    def inner_forward_impl(self, x: Tensor, quant_weight: Tensor, quant_bias: Optional[Tensor]):
        if self.is_same_padded_strided:
            return self.conv1d_same_zeros_pad_stride(x, quant_weight, quant_bias)
        else:
            return self._conv_forward(x, quant_weight, quant_bias)


class QuantConv2d(QuantWBIOL, Conv2d):

    def __init__(
            self,
            in_channels: int,
            out_channels: int,
            kernel_size: Union[int, Tuple[int, int]],
            stride: Union[int, Tuple[int, int]] = 1,
            padding: Union[int, Tuple[int, int]] = 0,
            dilation: Union[int, Tuple[int, int]] = 1,
            groups: int = 1,
            padding_mode: str = 'zeros',
            bias: Optional[bool] = True,
            weight_quant: Optional[WeightQuantType] = Int8WeightPerTensorFloat,
            bias_quant: Optional[BiasQuantType] = None,
            input_quant: Optional[ActQuantType] = None,
            output_quant: Optional[ActQuantType] = None,
            return_quant_tensor: bool = False,
            device: Optional[torch.device] = None,
            dtype: Optional[torch.dtype] = None,
            **kwargs) -> None:
        # avoid an init error in the super class by setting padding to 0
        if padding_mode == 'zeros' and padding == 'same' and (stride > 1 if isinstance(
                stride, int) else any(map(lambda x: x > 1, stride))):
            padding = 0
            is_same_padded_strided = True
        else:
            is_same_padded_strided = False
        Conv2d.__init__(
            self,
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            padding_mode=padding_mode,
            dilation=dilation,
            groups=groups,
            bias=bias,
            device=device,
            dtype=dtype)
        QuantWBIOL.__init__(
            self,
            weight_quant=weight_quant,
            bias_quant=bias_quant,
            input_quant=input_quant,
            output_quant=output_quant,
            return_quant_tensor=return_quant_tensor,
            **kwargs)
        self.is_same_padded_strided = is_same_padded_strided

    @property
    def per_elem_ops(self):
        flat_kernel_size = self.kernel_size[0] * self.kernel_size[1]
        return 2 * flat_kernel_size * (self.in_channels // self.groups)

    @property
    def output_channel_dim(self):
        if self.transposed:
            return 1
        else:
            return 0

    @property
    def channelwise_separable(self) -> bool:
        return self.groups == self.in_channels

    def conv2d_same_zeros_pad_stride(self, x: Tensor, weight: Tensor, bias: Optional[Tensor]):
        ih, iw = x.size()[-2:]
        kh, kw = weight.size()[-2:]
        sh, sw = self.stride
        oh, ow = math.ceil(ih / sh), math.ceil(iw / sw)
        pad_h = max((oh - 1) * self.stride[0] + (kh - 1) * self.dilation[0] + 1 - ih, 0)
        pad_w = max((ow - 1) * self.stride[1] + (kw - 1) * self.dilation[1] + 1 - iw, 0)
        if pad_h > 0 or pad_w > 0:
            x = F.pad(x, [pad_w // 2, pad_w - pad_w // 2, pad_h // 2, pad_h - pad_h // 2])
        out = F.conv2d(x, weight, bias, self.stride, 0, self.dilation, self.groups)
        return out

    def forward(self, input: Union[Tensor, QuantTensor]) -> Union[Tensor, QuantTensor]:
        return self.forward_impl(input)

    def inner_forward_impl(self, x: Tensor, quant_weight: Tensor, quant_bias: Optional[Tensor]):
        if self.is_same_padded_strided:
            return self.conv2d_same_zeros_pad_stride(x, quant_weight, quant_bias)
        else:
            return self._conv_forward(x, quant_weight, quant_bias)


class QuantConv3d(QuantWBIOL, Conv3d):

    def __init__(
            self,
            in_channels: int,
            out_channels: int,
            kernel_size: Union[int, Tuple[int, int]],
            stride: Union[int, Tuple[int, int]] = 1,
            padding: Union[int, Tuple[int, int]] = 0,
            dilation: Union[int, Tuple[int, int]] = 1,
            groups: int = 1,
            padding_mode: str = 'zeros',
            bias: bool = True,
            weight_quant: Optional[WeightQuantType] = Int8WeightPerTensorFloat,
            bias_quant: Optional[BiasQuantType] = None,
            input_quant: Optional[ActQuantType] = None,
            output_quant: Optional[ActQuantType] = None,
            return_quant_tensor: bool = False,
            device: Optional[torch.device] = None,
            dtype: Optional[torch.dtype] = None,
            **kwargs) -> None:
        # avoid an init error in the super class by setting padding to 0
        if padding_mode == 'zeros' and padding == 'same' and (stride > 1 if isinstance(
                stride, int) else any(map(lambda x: x > 1, stride))):
            padding = 0
            is_same_padded_strided = True
        else:
            is_same_padded_strided = False
        Conv3d.__init__(
            self,
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            dilation=dilation,
            groups=groups,
            padding_mode=padding_mode,
            bias=bias,
            device=device,
            dtype=dtype)
        QuantWBIOL.__init__(
            self,
            weight_quant=weight_quant,
            bias_quant=bias_quant,
            input_quant=input_quant,
            output_quant=output_quant,
            return_quant_tensor=return_quant_tensor,
            **kwargs)
        self.is_same_padded_strided = is_same_padded_strided

    @property
    def per_elem_ops(self):
        flat_kernel_size = self.kernel_size[0] * self.kernel_size[1] * self.kernel_size[2]
        return 2 * flat_kernel_size * (self.in_channels // self.groups)

    @property
    def output_channel_dim(self):
        if self.transposed:
            return 1
        else:
            return 0

    @property
    def channelwise_separable(self) -> bool:
        # according to https://pytorch.org/docs/stable/generated/torch.nn.Conv3d.html
        # if groups == in_channels that means each channel is convolved with its own set of filters
        return self.groups == self.channels

    def conv3d_same_zeros_pad_stride(self, x, weight, bias):
        id, ih, iw = x.size()[-3:]
        kd, kh, kw = weight.size()[-3:]
        sd, sh, sw = self.stride
        od, oh, ow = math.ceil(id / sd), math.ceil(ih / sh), math.ceil(iw / sw)
        pad_d = max((od - 1) * self.stride[0] + (kd - 1) * self.dilation[0] + 1 - id, 0)
        pad_h = max((oh - 1) * self.stride[1] + (kh - 1) * self.dilation[1] + 1 - ih, 0)
        pad_w = max((ow - 1) * self.stride[2] + (kw - 1) * self.dilation[2] + 1 - iw, 0)
        if pad_h > 0 or pad_w > 0:
            x = F.pad(
                x,
                [
                    pad_w // 2,
                    pad_w - pad_w // 2,
                    pad_h // 2,
                    pad_h - pad_h // 2,
                    pad_d // 2,
                    pad_d - pad_d // 2])
        out = F.conv3d(x, weight, bias, self.stride, 0, self.dilation, self.groups)
        return out

    def forward(self, input: Union[Tensor, QuantTensor]) -> Union[Tensor, QuantTensor]:
        # calls QuantWBIOL.forward_impl and eventually inner_forward_impl below
        return self.forward_impl(input)

    # override of QuantWBIOL method, called by QuantWBIOL.forward_impl
    def inner_forward_impl(self, x: Tensor, quant_weight: Tensor, quant_bias: Optional[Tensor]):
        if self.is_same_padded_strided:
            return self.conv3d_same_zeros_pad_stride(x, quant_weight, quant_bias)
        else:
            return self._conv_forward(x, quant_weight, quant_bias)



"""
Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
"""
import re

from torch import nn

from brevitas import nn as qnn
from brevitas.core.zero_point import ParameterFromStatsFromParameterZeroPoint
from brevitas.graph.quantize import layerwise_quantize
from brevitas.quant.experimental.float import Fp8e4m3Act
from brevitas.quant.experimental.float import Fp8e4m3ActPerTensorFloat
from brevitas.quant.experimental.float import Fp8e4m3WeightPerChannelFloat
from brevitas.quant.experimental.float import Fp8e4m3WeightPerTensorFloat
from brevitas.quant.fixed_point import Int8ActPerTensorFixedPoint
from brevitas.quant.fixed_point import Int8ActPerTensorFixedPointMSE
from brevitas.quant.fixed_point import Int8WeightPerChannelFixedPoint
from brevitas.quant.fixed_point import Int8WeightPerChannelFixedPointMSE
from brevitas.quant.fixed_point import Int8WeightPerTensorFixedPoint
from brevitas.quant.fixed_point import Int8WeightPerTensorFixedPointMSE
from brevitas.quant.scaled_int import Int8ActPerTensorFloat
from brevitas.quant.scaled_int import Int8ActPerTensorFloatMSE
from brevitas.quant.scaled_int import Int8WeightPerChannelFloat
from brevitas.quant.scaled_int import Int8WeightPerChannelFloatMSE
from brevitas.quant.scaled_int import Int8WeightPerTensorFloat
from brevitas.quant.scaled_int import Int8WeightPerTensorFloatMSE
from brevitas.quant.shifted_scaled_int import ShiftedUint8ActPerTensorFloat
from brevitas.quant.shifted_scaled_int import ShiftedUint8ActPerTensorFloatMSE
from brevitas.quant.shifted_scaled_int import ShiftedUint8WeightPerChannelFloat
from brevitas.quant.shifted_scaled_int import ShiftedUint8WeightPerChannelFloatMSE
from brevitas.quant.shifted_scaled_int import ShiftedUint8WeightPerTensorFloat
from brevitas.quant.shifted_scaled_int import ShiftedUint8WeightPerTensorFloatMSE
from brevitas_examples.llm.llm_quant.quantizers import Fp8e4m3WeightSymmetricGroupQuant
from brevitas_examples.llm.llm_quant.quantizers import Int8ActDynamicPerGroupFloat
from brevitas_examples.llm.llm_quant.quantizers import Int8ActDynamicPerRowFloat
from brevitas_examples.llm.llm_quant.quantizers import Int8ActDynamicPerTensorFloat
from brevitas_examples.llm.llm_quant.quantizers import Int8ActPerRowFloat
from brevitas_examples.llm.llm_quant.quantizers import Int8ActPerRowFloatMSE
from brevitas_examples.llm.llm_quant.quantizers import IntWeightSymmetricGroupQuant
from brevitas_examples.llm.llm_quant.quantizers import ShiftedUint8ActPerRowFloat
from brevitas_examples.llm.llm_quant.quantizers import ShiftedUint8ActPerRowFloatMSE
from brevitas_examples.llm.llm_quant.quantizers import ShiftedUintWeightAsymmetricGroupQuant

WEIGHT_QUANT_MAP = {
    'int': {
        'float_scale': {
            'stats': {
                'per_tensor': {
                    'sym': Int8WeightPerTensorFloat, 'asym': ShiftedUint8WeightPerTensorFloat},
                'per_channel': {
                    'sym': Int8WeightPerChannelFloat, 'asym': ShiftedUint8WeightPerChannelFloat},
                'per_group': {
                    'sym': IntWeightSymmetricGroupQuant,
                    'asym': ShiftedUintWeightAsymmetricGroupQuant},},
            'mse': {
                'per_tensor': {
                    'sym': Int8WeightPerTensorFloatMSE,
                    'asym': ShiftedUint8WeightPerTensorFloatMSE},
                'per_channel': {
                    'sym': Int8WeightPerChannelFloatMSE,
                    'asym': ShiftedUint8WeightPerChannelFloatMSE},},},
        'po2_scale': {
            'stats': {
                'per_tensor': {
                    'sym': Int8WeightPerTensorFixedPoint},
                'per_channel': {
                    'sym': Int8WeightPerChannelFixedPoint},},
            'mse': {
                'per_tensor': {
                    'sym': Int8WeightPerTensorFixedPointMSE},
                'per_channel': {
                    'sym': Int8WeightPerChannelFixedPointMSE},},}},
    'float': {
        'float_scale': {
            'stats': {
                'per_tensor': {
                    'sym': Fp8e4m3WeightPerTensorFloat},
                'per_channel': {
                    'sym': Fp8e4m3WeightPerChannelFloat},
                'per_group': {
                    'sym': Fp8e4m3WeightSymmetricGroupQuant}},}}}

INPUT_QUANT_MAP = {
    'int': {
        'static': {
            'float_scale': {
                'stats': {
                    'per_tensor': {
                        'sym': Int8ActPerTensorFloat, 'asym': ShiftedUint8ActPerTensorFloat},
                    'per_row': {
                        'sym': Int8ActPerRowFloat, 'asym': ShiftedUint8ActPerRowFloat},},
                'mse': {
                    'per_tensor': {
                        'sym': Int8ActPerTensorFloatMSE, 'asym': ShiftedUint8ActPerTensorFloatMSE},
                    'per_row': {
                        'sym': Int8ActPerRowFloatMSE, 'asym': ShiftedUint8ActPerRowFloatMSE},},},
            'po2_scale': {
                'stats': {
                    'per_tensor': {
                        'sym': Int8ActPerTensorFixedPoint},},
                'mse': {
                    'per_tensor': {
                        'sym': Int8ActPerTensorFixedPointMSE},},}},
        'dynamic': {
            'float_scale': {
                'stats': {
                    'per_tensor': {
                        'sym': Int8ActDynamicPerTensorFloat},
                    'per_row': {
                        'sym': Int8ActDynamicPerRowFloat},
                    'per_group': {
                        'sym': Int8ActDynamicPerGroupFloat},}}}},
    'float': {
        'static': {
            'float_scale': {
                'stats': {
                    'per_tensor': {
                        'sym': Fp8e4m3ActPerTensorFloat},}}},
        'no_scale': {
            'sym': Fp8e4m3Act,}}}


def quantize_model(
        model,
        dtype,
        weight_bit_width,
        weight_param_method,
        weight_scale_precision,
        weight_quant_type,
        weight_quant_granularity,
        weight_group_size,
        quantize_weight_zero_point,
        weight_quant_format='int',
        input_bit_width=None,
        input_quant_format=None,
        input_scale_precision=None,
        input_scale_type=None,
        input_param_method=None,
        input_quant_type=None,
        input_quant_granularity=None,
        input_group_size=None,
        quantize_input_zero_point=False,
        quantize_embedding=False,
        seqlen=None):
    """
    Replace float layers with quant layers in the target model
    """
    # Retrive base input and weight quantizers

    # match against custom float format
    if re.compile(r'e[1-8]m[1-8]').match(weight_quant_format):
        weight_float_format = {
            'exponent_bit_width': int(weight_quant_format[1]),
            'mantissa_bit_width': int(weight_quant_format[3])}
        weight_quant_format = 'float'
    else:
        weight_float_format = {}
    if input_quant_format is not None and re.compile(r'e[1-8]m[1-8]').match(input_quant_format):
        input_float_format = {
            'exponent_bit_width': int(input_quant_format[1]),
            'mantissa_bit_width': int(input_quant_format[3])}
        input_quant_format = 'float'
    else:
        input_float_format = {}

    weight_quant = WEIGHT_QUANT_MAP[weight_quant_format][weight_scale_precision][
        weight_param_method][weight_quant_granularity][weight_quant_type]
    if input_bit_width is not None and input_scale_type == 'no_scale':
        input_quant = sym_input_quant = linear_2d_input_quant = INPUT_QUANT_MAP[input_quant_format][
            input_scale_type][input_quant_type]
    elif input_bit_width is not None:
        input_quant = INPUT_QUANT_MAP[input_quant_format][input_scale_type][input_scale_precision][
            input_param_method][input_quant_granularity][input_quant_type]
        # Some activations in MHA should always be symmetric
        sym_input_quant = INPUT_QUANT_MAP[input_quant_format][input_scale_type][
            input_scale_precision][input_param_method][input_quant_granularity]['sym']
        # Linear layers with 2d input should always be per tensor or per group, as there is no row dimension
        if input_quant_granularity == 'per_tensor' or input_quant_granularity == 'per_row':
            linear_2d_input_quant = INPUT_QUANT_MAP[input_quant_format][input_scale_type][
                input_scale_precision][input_param_method]['per_tensor'][input_quant_type]
        else:
            assert input_quant_granularity == 'per_group'
            linear_2d_input_quant = input_quant
    else:
        input_quant = None
        sym_input_quant = None
        linear_2d_input_quant = None

    # Modify the weight quantizer based on the arguments passed in
    weight_quant = weight_quant.let(
        **{
            'bit_width': weight_bit_width,
            'narrow_range': False,
            'block_size': weight_group_size,
            'quantize_zero_point': quantize_weight_zero_point},
        **weight_float_format)
    # weight scale is converted to a standalone parameter
    # This is done already by default in the per_group quantizer
    if weight_quant_granularity != 'per_group':
        weight_quant = weight_quant.let(scaling_impl_type='parameter_from_stats')
    # weight zero-point is converted to a standalone parameter
    # This is done already by default in the per_group quantizer
    if weight_quant_type == 'asym' and weight_quant_granularity != 'per_group':
        weight_quant = weight_quant.let(zero_point_impl=ParameterFromStatsFromParameterZeroPoint)

    # Modify the input quantizers based on the arguments passed in
    if input_quant is not None:
        input_quant = input_quant.let(
            **{
                'bit_width': input_bit_width,
                'quantize_zero_point': quantize_input_zero_point,
                'dtype': dtype,},
            **input_float_format)
        if input_scale_type == 'static' and input_quant_granularity == 'per_row':
            # QuantMHA internally always uses Seq, B, E
            input_quant = input_quant.let(
                **{
                    'per_channel_broadcastable_shape': (seqlen, 1, 1),
                    'scaling_stats_permute_dims': (0, 1, 2)})
        elif input_scale_type == 'dynamic':
            if input_quant_granularity == 'per_tensor':
                input_quant = input_quant.let(
                    **{
                        'dynamic_scaling_broadcastable_shape': (1, -1, 1),
                        'permute_dims': (1, 0, 2),
                        'stats_reduce_dim': 1})
            elif input_quant_granularity == 'per_row':
                input_quant = input_quant.let(
                    **{
                        'dynamic_scaling_broadcastable_shape': (seqlen, -1, 1),
                        'permute_dims': (1, 0, 2),
                        'stats_reduce_dim': 2})
            elif input_quant_granularity == 'per_group':
                input_quant = input_quant.let(**{'group_dim': 2, 'group_size': input_group_size})
    if sym_input_quant is not None:
        sym_input_quant = sym_input_quant.let(
            **{
                'bit_width': input_bit_width,
                'quantize_zero_point': quantize_input_zero_point,
                'dtype': dtype},
            **input_float_format)
        if input_scale_type == 'static' and input_quant_granularity == 'per_row':
            q_scaled_quant = sym_input_quant.let(
                **{
                    'per_channel_broadcastable_shape': (1, seqlen, 1),
                    'scaling_stats_permute_dims': (1, 0, 2)})
            k_transposed_quant = sym_input_quant.let(
                **{
                    'per_channel_broadcastable_shape': (1, 1, seqlen),
                    'scaling_stats_permute_dims': (2, 0, 1)})
            v_quant = q_scaled_quant
            attn_output_weights_quant = q_scaled_quant
        elif input_scale_type == 'dynamic':
            if input_quant_granularity == 'per_tensor':
                q_scaled_quant = sym_input_quant.let(
                    **{
                        'dynamic_scaling_broadcastable_shape': (-1, 1, 1),
                        'permute_dims': None,
                        'stats_reduce_dim': 1})
                k_transposed_quant = sym_input_quant.let(
                    **{
                        'dynamic_scaling_broadcastable_shape': (-1, 1, 1),
                        'permute_dims': None,
                        'stats_reduce_dim': 1})
            elif input_quant_granularity == 'per_row':
                q_scaled_quant = sym_input_quant.let(
                    **{
                        'dynamic_scaling_broadcastable_shape': (-1, seqlen, 1),
                        'permute_dims': None,
                        'stats_reduce_dim': 2})
                k_transposed_quant = sym_input_quant.let(
                    **{
                        'dynamic_scaling_broadcastable_shape': (-1, 1, seqlen),
                        'permute_dims': None,
                        'stats_reduce_dim': 1})
            elif input_quant_granularity == 'per_group':
                q_scaled_quant = sym_input_quant.let(
                    **{
                        'group_dim': 2, 'group_size': input_group_size})
                k_transposed_quant = sym_input_quant.let(
                    **{
                        'group_dim': 1, 'group_size': input_group_size})
            v_quant = q_scaled_quant
            attn_output_weights_quant = q_scaled_quant
        else:
            q_scaled_quant = v_quant = k_transposed_quant = attn_output_weights_quant = sym_input_quant
    else:
        q_scaled_quant = v_quant = k_transposed_quant = attn_output_weights_quant = None
    if linear_2d_input_quant is not None:
        linear_2d_input_quant = linear_2d_input_quant.let(
            **{
                'bit_width': input_bit_width,
                'quantize_zero_point': quantize_input_zero_point,
                'dtype': dtype},
            **input_float_format)
        if input_scale_type == 'dynamic':
            # Note: this breaks if applied to 3d Linear inputs,
            # in case standard MHA layers haven't been inserted
            if input_quant_granularity == 'per_tensor' or input_quant_granularity == 'per_row':
                linear_2d_input_quant = linear_2d_input_quant.let(
                    **{
                        'dynamic_scaling_broadcastable_shape': (-1, 1),
                        'permute_dims': None,
                        'stats_reduce_dim': 1})
            elif input_quant_granularity == 'per_group':
                linear_2d_input_quant = linear_2d_input_quant.let(
                    **{
                        'group_dim': 1, 'group_size': input_group_size})

    quant_linear_kwargs = {
        'input_quant': linear_2d_input_quant, 'weight_quant': weight_quant, 'dtype': dtype}

    quant_mha_kwargs = {
        'in_proj_input_quant': input_quant,
        'in_proj_weight_quant': weight_quant,
        'in_proj_bias_quant': None,
        'softmax_input_quant': None,
        'attn_output_weights_quant': attn_output_weights_quant,
        'attn_output_weights_signed': input_quant_format == 'float',
        'q_scaled_quant': q_scaled_quant,
        'k_transposed_quant': k_transposed_quant,
        'v_quant': v_quant,
        'out_proj_input_quant': input_quant,
        'out_proj_weight_quant': weight_quant,
        'out_proj_bias_quant': None,
        'out_proj_output_quant': None,
        'batch_first': True,
        # activation equalization requires packed_in_proj
        # since it supports only self-attention
        'packed_in_proj': True,
        'dtype': dtype}

    layer_map = {
        nn.Linear: (qnn.QuantLinear, quant_linear_kwargs),
        nn.MultiheadAttention: (qnn.QuantMultiheadAttention, quant_mha_kwargs)}

    if quantize_embedding:
        quant_embedding_kwargs = {'weight_quant': weight_quant, 'dtype': dtype}
        layer_map[nn.Embedding] = (qnn.QuantEmbedding, quant_embedding_kwargs)

    layerwise_quantize(model=model, compute_layer_map=layer_map)



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

import math
from typing import List

from dependencies import this
from dependencies import value
import torch
from torch import Tensor

from brevitas.core.bit_width import *
from brevitas.core.function_wrapper import TensorClamp
from brevitas.core.function_wrapper import TensorClampSte
from brevitas.core.function_wrapper.misc import Identity
from brevitas.core.function_wrapper.shape import StatsInputViewShapeImpl
from brevitas.core.scaling import *
from brevitas.core.scaling import ScalingImplType
from brevitas.core.scaling import ScalingPerOutputType
from brevitas.inject import ExtendedInjector
from brevitas.quant.solver.common import *

__all__ = [
    'ScalingConstInit',
    'HeScalingInit',
    'SolveParameterTensorClampImplFromEnum',
    'SolveParameterScalingInitFromEnum',
    'SolveParameterScalingImplFromEnum',
    'SolveParameterScalingShape']


class ScalingConstInit:

    def __init__(self, scaling_const):
        self.scaling_const = scaling_const

    def __call__(self):
        return self.scaling_const


class HeScalingInit:

    def __init__(self, tracked_parameter_list: List[torch.nn.Parameter]):
        self.tracked_parameter_list = tracked_parameter_list

    def __call__(self):
        scaling_init = 0.0
        # takes average of He scaling over parameter list
        for param in self.tracked_parameter_list:
            two_dim_param = param.view(param.shape[0], -1)
            scaling_init += math.sqrt(2.0 / two_dim_param.shape[1])
        scaling_init /= len(self.tracked_parameter_list)
        return torch.tensor(scaling_init)


class SolveParameterTensorClampImplFromEnum(ExtendedInjector):

    @value
    def tensor_clamp_impl(bit_width_impl_type, scaling_impl_type):
        if (bit_width_impl_type == BitWidthImplType.PARAMETER or
                scaling_impl_type == ScalingImplType.AFFINE_STATS or
                scaling_impl_type == ScalingImplType.PARAMETER_FROM_STATS or
                scaling_impl_type == ScalingImplType.PARAMETER):
            return TensorClamp
        else:
            return TensorClampSte


class SolveParameterScalingInitFromEnum(ExtendedInjector):

    @value
    def scaling_init(scaling_init_impl):
        scaling_init = scaling_init_impl()
        if isinstance(scaling_init, Tensor):
            return scaling_init.detach()
        else:
            return torch.tensor(scaling_init)

    @value
    def scaling_init_impl(scaling_impl_type):
        if scaling_impl_type == ScalingImplType.CONST:
            return ScalingConstInit
        elif scaling_impl_type == ScalingImplType.PARAMETER:
            return ScalingConstInit
        elif scaling_impl_type == ScalingImplType.HE:
            return HeScalingInit
        else:
            return None


class SolveParameterScalingImplFromEnum(SolveAffineRescalingFromEnum):

    @value
    def scaling_impl(scaling_impl_type):
        if scaling_impl_type == ScalingImplType.PARAMETER:
            return ParameterScaling
        elif scaling_impl_type == ScalingImplType.PARAMETER_FROM_STATS:
            return ParameterFromStatsFromParameterScaling
        elif scaling_impl_type == ScalingImplType.CONST:
            return ConstScaling
        elif scaling_impl_type == ScalingImplType.HE:
            return ConstScaling
        elif scaling_impl_type == ScalingImplType.STATS:
            return StatsFromParameterScaling
        elif scaling_impl_type == ScalingImplType.AFFINE_STATS:
            return StatsFromParameterScaling
        else:
            raise RuntimeError(f"{scaling_impl_type} not recognized.")


class SolveParameterScalingShape(ExtendedInjector):

    @value
    def scaling_shape(scaling_per_output, expanded_groupwise_shape=None, group_dim=None):
        if scaling_per_output == ScalingPerOutputType.TENSOR:
            return SCALAR_SHAPE
        elif scaling_per_output == ScalingPerOutputType.CHANNEL:
            return this.scaling_per_output_channel_shape
        elif scaling_per_output == ScalingPerOutputType.GROUP:
            # Scaling shape is like expanded_groupwise_shape but has 1 in position group_dim + 1
            assert expanded_groupwise_shape is not None, "Per Group scaling not correctly configured"
            assert group_dim is not None, "Per Group scaling not correctly configured"
            size = list(expanded_groupwise_shape)
            size[group_dim + 1] = 1
            return tuple(size)

    @value
    def reshaped_groupwise_shape(expanded_groupwise_shape, group_dim, group_size):
        new_shape = list(expanded_groupwise_shape)
        del new_shape[group_dim + 1]  # delete the group_size shape
        # Expand the group_dim shape, accounting for padding
        new_shape[group_dim] = new_shape[group_dim] * group_size
        return new_shape

    @value
    def expanded_groupwise_shape(tracked_parameter_list, group_dim, group_size=None):
        # expanded_groupwise_shape will be called always to create scaling_shape, but it is only needed
        # for groupwise quantization. All other groupwise shape infos are derived from this.

        # If conditions do not allow for groupwise quantization, early exit and return None
        if group_size is None:
            return

        # If group_size is specified and shared quantization is used, raise an error.
        assert len(tracked_parameter_list) == 1, "Shared groupwise quantization is not currently supported"

        weight_shape = tracked_parameter_list[0].shape
        size = list(weight_shape)
        size[group_dim] = (size[group_dim] + group_size - 1) // group_size
        size.insert(group_dim + 1, group_size)
        return tuple(size)

    @value
    def group_dim(module, group_size=None):
        # group_dim will be called always to create scaling_shape, but it is only needed
        # for groupwise quantization.
        if group_size is not None:
            return 1 if not hasattr(module, 'transposed') or not module.transposed else 0


class SolveInputViewImpl(ExtendedInjector):

    @value
    def input_view_impl(scaling_per_output):
        if scaling_per_output == ScalingPerOutputType.GROUP:
            return StatsInputViewShapeImpl.OVER_SUBCHANNEL_BLOCK
        else:
            return Identity



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

from typing import Optional

import torch
from torch import Tensor
from torch.nn import Module

import brevitas
import brevitas.config as config
from brevitas.core.utils import StatelessBuffer
from brevitas.function.ops_ste import tensor_clamp_ste


class BitWidthConst(brevitas.jit.ScriptModule):
    """
    ScriptModule that returns a constant bit-width wrapped in a float torch.tensor.

    Args:
        bit_width (int): bit-width value.

    Examples:
        >>> bit_width = BitWidthConst(8)
        >>> bit_width()
        tensor(8.)

    Note:
        The bit-width is not part of the Module's state, meaning that it won't be saved as part of
        a checkpoint.

    Note:
        Maps to bit_width_impl_type == BitWidthImplType.CONST == 'CONST' == 'const' in higher-level APIs.
    """

    def __init__(
            self,
            bit_width: int,
            dtype: Optional[torch.dtype] = None,
            device: Optional[torch.device] = None) -> None:
        super(BitWidthConst, self).__init__()
        assert isinstance(bit_width, int)
        self.bit_width = StatelessBuffer(torch.tensor(float(bit_width), dtype=dtype, device=device))

    @brevitas.jit.script_method
    def forward(self) -> Tensor:
        return self.bit_width()


class BitWidthStatefulConst(brevitas.jit.ScriptModule):
    """
    ScriptModule that returns a constant bit-width wrapped in a float torch.tensor but retains the
    bit-width as part of the module state.

    Args:
        bit_width (int): bit-width value.

    Examples:
        >>> bit_width = BitWidthStatefulConst(8)
        >>> bit_width()
        tensor(8.)

    Note:
        The BitWidthStatefulConst is a counterpart to BitWidthConst with the difference that the
        BitWidthStatefulConst retains the bit-width as part of the Module's state. This means that it
        will be saved as part of a checkpoint.

    Note:
        Maps to bit_width_impl_type == BitWidthImplType.STATEFUL_CONST == 'STATEFUL_CONST' ==
        'stateful_const' in higher-level APIs.
    """

    def __init__(
            self,
            bit_width: int,
            dtype: Optional[torch.dtype] = None,
            device: Optional[torch.device] = None) -> None:
        super(BitWidthStatefulConst, self).__init__()
        assert isinstance(bit_width, int)
        self.register_buffer(
            "bit_width", torch.tensor(float(bit_width), dtype=dtype, device=device))

    @brevitas.jit.script_method
    def forward(self) -> Tensor:
        return self.bit_width

    def _load_from_state_dict(
            self,
            state_dict,
            prefix,
            local_metadata,
            strict,
            missing_keys: list,
            unexpected_keys,
            error_msgs):
        super(BitWidthStatefulConst, self)._load_from_state_dict(
            state_dict, prefix, local_metadata, strict, missing_keys, unexpected_keys, error_msgs)
        value_key = prefix + "bit_width"
        if config.IGNORE_MISSING_KEYS and value_key in missing_keys:
            missing_keys.remove(value_key)


class MsbClampBitWidth(brevitas.jit.ScriptModule):

    def __init__(
            self,
            bit_width_to_remove_impl: Module,
            min_overall_bit_width: int,
            max_overall_bit_width: int) -> None:
        super(MsbClampBitWidth, self).__init__()

        self.min_overall_bit_width = BitWidthConst(min_overall_bit_width)
        self.max_overall_bit_width = BitWidthConst(max_overall_bit_width)
        self.bit_width_to_remove_impl = bit_width_to_remove_impl

    @brevitas.jit.script_method
    def forward(self, input_bit_width: Tensor) -> Tensor:
        bit_width_to_remove = self.bit_width_to_remove_impl()
        output_bit_width = torch.abs(input_bit_width - bit_width_to_remove)
        output_bit_width = tensor_clamp_ste(
            output_bit_width, self.min_overall_bit_width(), self.max_overall_bit_width())
        return output_bit_width



import functools

import torch
import torch.nn.functional as F

FLOAT_QUANT_TENSOR_FN_HANDLER = {}


def implements_float_qt(torch_function):

    @functools.wraps(torch_function)
    def decorator(func):
        FLOAT_QUANT_TENSOR_FN_HANDLER[torch_function] = func
        return func

    return decorator


@implements_float_qt(torch.cat)
def cat_handler(*args, **kwargs):
    from brevitas.quant_tensor import FloatQuantTensor
    return FloatQuantTensor.cat(*args, **kwargs)


@implements_float_qt(F.conv1d)
def conv1d_handler(quant_input, quant_weight, bias=None, *args, **kwargs):
    output = quant_layer(F.conv1d, quant_input, quant_weight, bias, *args, **kwargs)
    return output


@implements_float_qt(F.conv2d)
def conv2d_handler(quant_input, quant_weight, bias=None, *args, **kwargs):
    output = quant_layer(F.conv2d, quant_input, quant_weight, bias, *args, **kwargs)
    return output


@implements_float_qt(F.conv3d)
def conv3d_handler(quant_input, quant_weight, bias=None, *args, **kwargs):
    output = quant_layer(F.conv3d, quant_input, quant_weight, bias, *args, **kwargs)
    return output


@implements_float_qt(F.conv_transpose1d)
def conv_transpose1d_handler(quant_input, quant_weight, bias=None, *args, **kwargs):
    output = quant_layer(F.conv_transpose1d, quant_input, quant_weight, bias, *args, **kwargs)
    return output


@implements_float_qt(F.conv_transpose2d)
def conv_transpose2d_handler(quant_input, quant_weight, bias=None, *args, **kwargs):
    output = quant_layer(F.conv_transpose2d, quant_input, quant_weight, bias, *args, **kwargs)
    return output


@implements_float_qt(F.conv_transpose3d)
def conv_transpose3d_handler(quant_input, quant_weight, bias=None, *args, **kwargs):
    output = quant_layer(F.conv_transpose3d, quant_input, quant_weight, bias, *args, **kwargs)
    return output


@implements_float_qt(F.linear)
def linear_handler(quant_input, quant_weight, bias=None, *args, **kwargs):
    output = quant_layer(F.linear, quant_input, quant_weight, bias, *args, **kwargs)
    return output


@implements_float_qt(F.embedding)
def embedding_handler(input, quant_weight, *args, **kwargs):
    from brevitas.quant_tensor import _unpack_quant_tensor
    from brevitas.quant_tensor import FloatQuantTensor

    quant_weight_value = _unpack_quant_tensor(quant_weight)
    out = F.embedding(input, quant_weight_value, *args, **kwargs)

    if isinstance(quant_weight, FloatQuantTensor):
        scale = quant_weight.scale
        zero_point = quant_weight.zero_point
        exponent_bit_width = quant_weight.exponent_bit_width
        mantissa_bit_width = quant_weight.mantissa_bit_width
        exponent_bias = quant_weight.exponent_bias
        inf_values = quant_weight.inf_values
        nan_values = quant_weight.nan_values
        if any(t.numel() > 1 for t in [scale, zero_point]):
            raise RuntimeError("Only per-tensor quantization is supported.")
        signed = quant_weight.signed
        training = quant_weight.training
        saturating = quant_weight.saturating
        out = FloatQuantTensor(
            out,
            scale,
            zero_point,
            exponent_bit_width,
            mantissa_bit_width,
            exponent_bias,
            saturating,
            inf_values,
            nan_values,
            signed,
            training)
    return out


@implements_float_qt(F.avg_pool2d)
def avg_pool2d_handler(
        quant_input, kernel_size, stride, padding, ceil_mode, count_include_pad, divisor_override):
    from brevitas.quant_tensor import _unpack_quant_tensor

    x = F.avg_pool2d(
        _unpack_quant_tensor(quant_input),
        kernel_size,
        stride,
        padding,
        ceil_mode,
        count_include_pad,
        divisor_override)

    return x


@implements_float_qt(F.adaptive_avg_pool2d)
def adaptive_avg_pool2d_handler(quant_input, output_shape):
    from brevitas.quant_tensor import _unpack_quant_tensor

    x = F.adaptive_avg_pool2d(_unpack_quant_tensor(quant_input), output_shape)
    return x


def quant_layer(fn, quant_input, quant_weight, bias, *args, **kwargs):
    from brevitas.quant_tensor import _unpack_quant_tensor

    if bias is None:
        output = fn(
            _unpack_quant_tensor(quant_input),
            _unpack_quant_tensor(quant_weight),
            None,
            *args,
            **kwargs)
    else:
        output = fn(
            _unpack_quant_tensor(quant_input),
            _unpack_quant_tensor(quant_weight),
            _unpack_quant_tensor(bias),
            *args,
            **kwargs)

    return output



from ._symbolic_trace import PH
from ._symbolic_trace import ProxyableClassMeta
from ._symbolic_trace import symbolic_trace
from ._symbolic_trace import Tracer
from ._symbolic_trace import wrap
from .graph import CodeGen
from .graph import Graph
from .graph_module import GraphModule
from .interpreter import Interpreter as Interpreter
from .interpreter import Transformer as Transformer
from .node import map_arg
from .node import Node
from .proxy import Proxy
from .subgraph_rewriter import replace_pattern



"""
Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

Adapted from https://github.com/IST-DASLab/gptq, released under the following LICENSE:

Copyright 2023 IST-DASLab

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from contextlib import contextmanager
import inspect

from optimum.utils.normalized_config import NormalizedConfigManager
import torch
from torch.utils._python_dispatch import TorchDispatchMode
from torch.utils._pytree import tree_map
from transformers import AutoConfig
from transformers.utils.fx import symbolic_trace

from brevitas.fx.value_tracer import ValueProxy


def get_fx(model):
    forward_signature = inspect.signature(model.forward).parameters
    if all(input_name in forward_signature
           for input_name in ["input_ids", "attention_mask", "past_key_values"]):
        input_names = ["input_ids", "attention_mask", "past_key_values"]
    else:
        raise ValueError(
            f"Quantization with an FX graph is currently only supported for models taking `input_ids`, `attention_mask` and `past_key_values` as inputs. The model only has the following inputs: {forward_signature}"
        )

    with torch.no_grad():
        model = symbolic_trace(model, input_names)
    return model


def modify_dataloader(model_name_or_path, data, dtype):
    config = AutoConfig.from_pretrained(model_name_or_path)

    normalized_config_class = NormalizedConfigManager.get_normalized_config_class(config.model_type)
    normalized_config = normalized_config_class(config)

    num_heads = normalized_config.num_attention_heads
    head_dim = normalized_config.hidden_size // num_heads
    num_layers = normalized_config.num_layers

    for sample in data:
        sample["past_key_values"] = tuple((
            torch.zeros(1, num_heads, 0, head_dim, device=sample["input_ids"].device, dtype=dtype),
            torch.zeros(1, num_heads, 0, head_dim, device=sample["input_ids"].device, dtype=dtype),
        ) for _ in range(num_layers))
    return data


@contextmanager
def cast_to_float32(model, target_dtype):
    dtype_dict = {}
    for name, p in model.state_dict().items():
        # This allows to pick up duplicated parameters
        dtype_dict[name] = p.dtype
    if any(dtype != torch.float32 for dtype in dtype_dict.values()):
        model.to(dtype=torch.float32)
    try:
        yield model
    finally:
        for name, p in {**dict(model.named_parameters()), **dict(model.named_buffers())}.items():
            if name in dtype_dict:
                p.data = p.data.to(dtype_dict[name])
            else:
                # target_dtype covers any new tensors that might have been
                # introduced in the process (e.g. during equalization)
                p.data = p.data.to(target_dtype)


class CastFloat16ToFloat32(TorchDispatchMode):

    def cast_cpu_to(self, x, src_dtype, dest_dtype):
        # workaround for value_trace to avoid tracing through the ops below
        if issubclass(type(x), ValueProxy):
            t = x.value
        else:
            t = x
        if isinstance(t, torch.Tensor) and t.dtype == src_dtype and t.device == torch.device('cpu'):
            # Keep the casting out of place so that it's ephemeral
            return t.to(dest_dtype, non_blocking=True, copy=True)
        return x

    def cast_cpu_float32(self, t):
        return self.cast_cpu_to(t, src_dtype=torch.float16, dest_dtype=torch.float32)

    def __torch_dispatch__(self, func, types, args=(), kwargs=None):
        kwargs = kwargs if kwargs else {}
        args, kwargs = tree_map(self.cast_cpu_float32, (args, kwargs))
        out = func(*args, **kwargs)
        return out



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

from brevitas.core.bit_width import BitWidthParameter
from brevitas.core.function_wrapper import *
from brevitas.core.quant import RescalingIntQuant
from brevitas.inject.enum import FloatToIntImplType
from brevitas.quant_tensor import FloatQuantTensor
from brevitas.quant_tensor import GroupwiseFloatQuantTensor
from brevitas.quant_tensor import GroupwiseIntQuantTensor
from brevitas.quant_tensor import IntQuantTensor


class _CachedIO:

    def __init__(self, quant_tensor: IntQuantTensor, metadata_only: bool):
        self.shape = quant_tensor.value.shape
        if metadata_only:
            self.value = None
            self.quant_tensor = quant_tensor.set(value=None)
        else:
            self.quant_tensor = quant_tensor
            # torch.compile compatibility
            self.value = quant_tensor.value
        # torch.compile compatibility
        self.scale = quant_tensor.scale

    @property
    def zero_point(self):
        return self.quant_tensor.zero_point

    @property
    def bit_width(self):
        return self.quant_tensor.bit_width

    @property
    def signed(self):
        return self.quant_tensor.signed


class _CachedIOFloat:

    def __init__(self, quant_tensor: FloatQuantTensor, metadata_only: bool):
        self.shape = quant_tensor.value.shape
        if metadata_only:
            self.value = None
            self.quant_tensor = quant_tensor.set(value=None)
        else:
            self.quant_tensor = quant_tensor
            # torch.compile compatibility
            self.value = quant_tensor.value
        # torch.compile compatibility
        self.scale = quant_tensor.scale

    @property
    def zero_point(self):
        return self.quant_tensor.zero_point

    @property
    def exponent_bit_width(self):
        return self.quant_tensor.exponent_bit_width

    @property
    def mantissa_bit_width(self):
        return self.quant_tensor.mantissa_bit_width

    @property
    def exponent_bias(self):
        return self.quant_tensor.exponent_bias

    @property
    def saturating(self):
        return self.quant_tensor.saturating

    @property
    def inf_values(self):
        return self.quant_tensor.inf_values

    @property
    def nan_values(self):
        return self.quant_tensor.nan_values

    @property
    def signed(self):
        return self.quant_tensor.signed


class _CachedIOGroupwiseFloat:

    def __init__(self, quant_tensor: GroupwiseFloatQuantTensor, metadata_only: bool):
        self.shape = quant_tensor.value.shape
        if metadata_only:
            self.value = None
            self.quant_tensor = quant_tensor.set(value=None)
        else:
            self.quant_tensor = quant_tensor
            # torch.compile compatibility
            self.value = quant_tensor.value
        # torch.compile compatibility
        self.scale = quant_tensor.scale

    @property
    def zero_point(self):
        return self.quant_tensor.zero_point

    @property
    def exponent_bit_width(self):
        return self.quant_tensor.exponent_bit_width

    @property
    def mantissa_bit_width(self):
        return self.quant_tensor.mantissa_bit_width

    @property
    def exponent_bias(self):
        return self.quant_tensor.exponent_bias

    @property
    def saturating(self):
        return self.quant_tensor.saturating

    @property
    def inf_values(self):
        return self.quant_tensor.inf_values

    @property
    def nan_values(self):
        return self.quant_tensor.nan_values

    @property
    def signed(self):
        return self.quant_tensor.signed

    @property
    def group_size(self):
        return self.quant_tensor.group_size

    @property
    def group_dim(self):
        return self.quant_tensor.group_dim


class _CachedIOGroupwiseInt:

    def __init__(self, quant_tensor: GroupwiseIntQuantTensor, metadata_only: bool):
        self.shape = quant_tensor.value.shape
        if metadata_only:
            self.value = None
            self.quant_tensor = quant_tensor.set(value=None)
        else:
            self.quant_tensor = quant_tensor
            # torch.compile compatibility
            self.value = quant_tensor.value
        # torch.compile compatibility
        self.scale = quant_tensor.scale

    @property
    def zero_point(self):
        return self.quant_tensor.zero_point

    @property
    def bit_width(self):
        return self.quant_tensor.bit_width

    @property
    def signed(self):
        return self.quant_tensor.signed

    @property
    def group_size(self):
        return self.quant_tensor.group_size

    @property
    def group_dim(self):
        return self.quant_tensor.group_dim


def has_learned_weight_bit_width(module):
    from brevitas.proxy.parameter_quant import WeightQuantProxyFromInjector

    if isinstance(module, WeightQuantProxyFromInjector) \
            and isinstance(module.tensor_quant, RescalingIntQuant) \
            and isinstance(module.tensor_quant.msb_clamp_bit_width_impl,
                           BitWidthParameter):
        return True
    else:
        return False


def has_learned_activation_bit_width(module):
    from brevitas.proxy.runtime_quant import ActQuantProxyFromInjector
    from brevitas.proxy.runtime_quant import FusedActivationQuantProxy

    if isinstance(module, ActQuantProxyFromInjector) \
            and isinstance(module.fused_activation_quant_proxy, FusedActivationQuantProxy) \
            and isinstance(module.fused_activation_quant_proxy.tensor_quant, RescalingIntQuant) \
            and isinstance(module.fused_activation_quant_proxy.tensor_quant.msb_clamp_bit_width_impl,
                           BitWidthParameter):
        return True
    else:
        return False


def float_to_int_impl_to_enum(module):
    if isinstance(module, RoundSte):
        return FloatToIntImplType.ROUND
    elif isinstance(module, RoundToZeroSte):
        return FloatToIntImplType.ROUND_TO_ZERO
    elif isinstance(module, FloorSte):
        return FloatToIntImplType.FLOOR
    elif isinstance(module, CeilSte):
        return FloatToIntImplType.CEIL
    elif isinstance(module, DPURoundSte):
        return FloatToIntImplType.DPU
    elif isinstance(module, LearnedRoundSte):
        return FloatToIntImplType.LEARNED_ROUND
    elif isinstance(module, StochasticRoundSte):
        if module.deterministic_inference:
            return FloatToIntImplType.ROUND
        else:
            return FloatToIntImplType.STOCHASTIC_ROUND
    else:
        return None



# Copyright (C) 2024, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

import argparse
import copy
from hashlib import sha256
import os
import random

import numpy as np
import torch
import torch.nn as nn
import torch.optim as optim
import torch.optim.lr_scheduler as lrs

import brevitas.config as config
from brevitas.export import export_qonnx
import brevitas_examples.imagenet_classification.a2q.utils as utils

parser = argparse.ArgumentParser()
parser.add_argument(
    "--data-root", type=str, required=True, help="Directory where the dataset is stored.")
parser.add_argument(
    "--model-name",
    type=str,
    default="quant_resnet18_w4a4_a2q_32b",
    help="Name of model to train. Default: 'quant_resnet18_w4a4_a2q_32b'",
    choices=utils.model_impl.keys())
parser.add_argument(
    "--save-path",
    type=str,
    default="outputs/",
    help="Directory where to save checkpoints. Default: 'outputs/'")
parser.add_argument(
    "--num-workers",
    type=int,
    default=0,
    help="Number of workers for the dataloader to use. Default: 0")
parser.add_argument(
    "--pin-memory",
    action="store_true",
    default=False,
    help="If true, pin memory for the dataloader.")
parser.add_argument(
    "--batch-size-train",
    type=int,
    default=256,
    help="Batch size for the training dataloader. Default: 256")
parser.add_argument(
    "--batch-size-test",
    type=int,
    default=512,
    help="Batch size for the testing dataloader. Default: 512")
parser.add_argument(
    "--batch-size-calibration",
    type=int,
    default=256,
    help="Batch size for the calibration dataloader. Default: 256")
parser.add_argument(
    "--calibration-samples",
    type=int,
    default=1000,
    help="Number of samples to use for calibration. Default: 1000")
parser.add_argument(
    "--weight-decay",
    type=float,
    default=1e-5,
    help="Weight decay for the Adam optimizer. Default: 0.00001")
parser.add_argument(
    "--lr-init", type=float, default=1e-3, help="Initial learning rate. Default: 0.001")
parser.add_argument(
    "--lr-step-size",
    type=int,
    default=30,
    help="Step size for the learning rate scheduler. Default: 30")
parser.add_argument(
    "--lr-gamma",
    type=float,
    default=0.1,
    help="Default gamma for the learning rate scheduler. Default: 0.1")
parser.add_argument(
    "--total-epochs", type=int, default=90, help="Total epoch to train the model for. Default: 90")
parser.add_argument(
    "--from-float-checkpoint",
    action="store_true",
    default=False,
    help="If true, use a pre-trained floating-point checkpoint.")
parser.add_argument(
    "--save-torch-model",
    action="store_true",
    default=False,
    help="If true, save torch model to specified save path.")
parser.add_argument(
    "--apply-act-calibration",
    action="store_true",
    default=False,
    help="If true, apply activation calibration to the quantized model.")
parser.add_argument(
    "--apply-bias-correction",
    action="store_true",
    default=False,
    help="If true, apply bias correction to the quantized model.")
parser.add_argument(
    "--apply-ep-init",
    action="store_true",
    default=False,
    help="If true, apply EP-init to the quantized model.")
parser.add_argument(
    "--export-to-qonnx", action="store_true", default=False, help="If true, export model to QONNX.")

# ignore missing keys when loading pre-trained checkpoint
config.IGNORE_MISSING_KEYS = True

SEED = 0
random.seed(SEED)
np.random.seed(SEED)
torch.manual_seed(SEED)

# create a random input for graph tracing
random_inp = torch.randn(1, 3, 32, 32)

if __name__ == "__main__":

    args = parser.parse_args()

    config.JIT_ENABLED = not args.export_to_qonnx

    # Initialize dataloaders
    print(f"Loading CIFAR10 dataset from {args.data_root}...")
    trainloader, testloader = utils.get_cifar10_dataloaders(
        data_root=args.data_root,
        batch_size_train=args.batch_size_train,
        batch_size_test=args.batch_size_test,
        num_workers=args.num_workers,
        pin_memory=args.pin_memory)
    calibloader = utils.create_calibration_dataloader(
        dataset=trainloader.dataset,
        batch_size=args.batch_size_calibration,
        num_workers=args.num_workers,
        subset_size=args.calibration_samples)

    model = utils.get_model_by_name(
        args.model_name, init_from_float_checkpoint=args.from_float_checkpoint)
    criterion = nn.CrossEntropyLoss()
    optimizer = optim.SGD(
        utils.filter_params(model.named_parameters(), args.weight_decay),
        lr=args.lr_init,
        weight_decay=args.weight_decay)
    scheduler = lrs.StepLR(optimizer, step_size=args.lr_step_size, gamma=args.lr_gamma)

    # Calibrate the quant model on the calibration dataset
    if args.apply_ep_init:
        print("Applying EP-init:")
        model = utils.apply_ep_init(model, random_inp)

    # Calibrate the quant model on the calibration dataset
    if args.apply_act_calibration:
        print("Applying activation calibration:")
        utils.apply_act_calibrate(calibloader, model)

    if args.apply_bias_correction:
        print("Applying bias correction:")
        utils.apply_bias_correction(calibloader, model)

    best_top_1, best_weights = 0., copy.deepcopy(model.state_dict())
    for epoch in range(args.total_epochs):

        train_loss = utils.train_for_epoch(trainloader, model, criterion, optimizer)
        test_top_1, test_top_5, test_loss = utils.evaluate_topk_accuracies(testloader, model, criterion)
        scheduler.step()

        print(
            f"[Epoch {epoch:03d}]",
            f"train_loss={train_loss:.3f},",
            f"test_loss={test_loss:.3f},",
            f"test_top_1={test_top_1:.1%},",
            f"test_top_5={test_top_5:.1%}",
            sep=" ")

        if test_top_1 >= best_top_1:
            best_weights = copy.deepcopy(model.state_dict())
            best_top_1 = test_top_1

    model.load_state_dict(best_weights)
    top_1, top_5, loss = utils.evaluate_topk_accuracies(testloader, model, criterion)
    print(f"Final: top_1={top_1:.1%}, top_5={top_5:.1%}, loss={loss:.3f}")

    # save checkpoint
    os.makedirs(args.save_path, exist_ok=True)
    if args.save_torch_model:
        ckpt_path = f"{args.save_path}/{args.model_name}.pth"
        torch.save(best_weights, ckpt_path)
        with open(ckpt_path, "rb") as _file:
            bytes = _file.read()
            model_tag = sha256(bytes).hexdigest()[:8]
        new_ckpt_path = f"{args.save_path}/{args.model_name}-{model_tag}.pth"
        os.rename(ckpt_path, new_ckpt_path)
        print(f"Saved model checkpoint to {new_ckpt_path}")

    if args.export_to_qonnx:
        export_qonnx(
            model.cpu(),
            input_t=random_inp.cpu(),
            export_path=f"{args.save_path}/{args.model_name}-{model_tag}.onnx")



"""
Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
"""

import torch
from tqdm import tqdm

from brevitas.graph.equalize import activation_equalization_mode
from brevitas.graph.equalize import EqualizeGraph


@torch.no_grad()
def apply_act_equalization(model, act_equalization_type, dataloader, alpha=0.5):
    if act_equalization_type == 'layerwise':
        with activation_equalization_mode(model, alpha, add_mul_node=True, layerwise=True):
            for inps in tqdm(dataloader):
                model(**inps)

    elif act_equalization_type == 'fx':
        assert model is not None, "FX Model is required to perform FX SmoothQuant"
        with activation_equalization_mode(model,
                                          alpha,
                                          add_mul_node=False,
                                          layerwise=False,
                                          co_optimize_act_weights=True):
            for inps in tqdm(dataloader):
                model(**inps)

    else:
        raise RuntimeError(f"{act_equalization_type} not supported.")


@torch.no_grad()
def apply_weight_equalization(graph_model, scale_computation_type='range'):
    EqualizeGraph(scale_computation_type=scale_computation_type).apply(graph_model)



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

from packaging import version
import torch

from brevitas import torch_version


def clear_class_registry():
    # torch.jit.trace leaks memory, this should help
    # https://github.com/pytorch/pytorch/issues/86537
    # https://github.com/pytorch/pytorch/issues/35600
    torch._C._jit_clear_class_registry()
    torch.jit._recursive.concrete_type_store = torch.jit._recursive.ConcreteTypeStore()
    if torch_version >= version.parse('1.7.0'):
        torch.jit._state._script_classes.clear()



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

from typing import List, Optional, Tuple

import torch

import brevitas

VALUE_ATTR_NAME = 'value'


@torch.jit.ignore
def inplace_tensor_add(tensor: torch.Tensor, value: torch.Tensor) -> torch.Tensor:
    tensor.add_(value)
    return tensor


@torch.jit.ignore
def inplace_tensor_mul(tensor: torch.Tensor, value: torch.Tensor) -> torch.Tensor:
    tensor.mul_(value)
    return tensor


@torch.jit.ignore
def inplace_momentum_update(
        tensor: torch.Tensor,
        update: torch.Tensor,
        momentum: Optional[float],
        counter: int,
        new_counter: int) -> torch.Tensor:
    if momentum is None:
        tensor.mul_(counter / new_counter)
        tensor.add_(update / new_counter)
    else:
        tensor.mul_(1 - momentum)
        tensor.add_(momentum * update)
    return tensor


class StatelessBuffer(brevitas.jit.ScriptModule):

    def __init__(self, value: torch.Tensor):
        super(StatelessBuffer, self).__init__()
        self.register_buffer(VALUE_ATTR_NAME, value)

    @brevitas.jit.script_method
    def forward(self):
        return self.value.detach()

    def _load_from_state_dict(
            self, state_dict, prefix, local_metadata, strict, missing_keys, unexpected_keys,
            error_msgs):
        super(StatelessBuffer, self)._load_from_state_dict(
            state_dict, prefix, local_metadata, strict, missing_keys, unexpected_keys, error_msgs)
        value_key = prefix + VALUE_ATTR_NAME
        if value_key in missing_keys:
            missing_keys.remove(value_key)

    def state_dict(self, destination=None, prefix='', keep_vars=False):
        output_dict = super(StatelessBuffer, self).state_dict(
            destination=destination, prefix=prefix, keep_vars=keep_vars)
        del output_dict[prefix + VALUE_ATTR_NAME]
        return output_dict


class SingleArgStatelessBuffer(brevitas.jit.ScriptModule):

    def __init__(self, value: torch.Tensor):
        super(SingleArgStatelessBuffer, self).__init__()
        self.const = StatelessBuffer(torch.tensor(value))

    @brevitas.jit.script_method
    def forward(self, placeholder):
        return self.const()


class ParameterWrapper(brevitas.jit.ScriptModule):

    def __init__(self, value: torch.Tensor):
        super(ParameterWrapper, self).__init__()
        self.register_parameter(VALUE_ATTR_NAME, value)

    @brevitas.jit.script_method
    def forward(self):
        return self.value


class SliceTensor(brevitas.jit.ScriptModule):

    def __init__(self) -> None:
        super().__init__()
        self.subtensor_slice_list = brevitas.jit.Attribute([None], List[Optional[Tuple[int, int]]])

    @torch.jit.ignore
    def eager_forward(self, x: torch.Tensor) -> torch.Tensor:
        slices = tuple(slice(*s) if s is not None else slice(s) for s in self.subtensor_slice_list)
        x = x[slices]
        return x

    @brevitas.jit.script_method
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        if torch.jit.is_scripting():
            for i, s in enumerate(self.subtensor_slice_list):
                if s is not None:
                    x = x.slice(i, s[0], s[1])
        else:
            x = self.eager_forward(x)
        return x



"""
Adapted from https://github.com/huggingface/optimum-amd, released under the following LICENSE:

MIT License

Copyright (c) 2023 Hugging Face

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import random
from typing import Any, Optional, Union

import numpy as np
from optimum.amd.brevitas.data_utils import DatasetToDevice
from optimum.amd.brevitas.data_utils import get_c4
from optimum.amd.brevitas.data_utils import get_wikitext2
from optimum.utils.normalized_config import NormalizedConfigManager
import torch
from transformers import AutoConfig


def get_dataset_for_model(
    model_name_or_path: str,
    dataset_name: str,
    tokenizer: Any,
    nsamples: int = 128,
    seqlen: int = 2048,
    seed: int = 0,
    split: str = "train",
    fuse_sequences: bool = True,
    require_fx: bool = False,
    device: Optional[Union[str, torch.device]] = None,
):
    random.seed(seed)
    np.random.seed(seed)
    torch.random.manual_seed(seed)
    get_dataset_map = {
        "wikitext2": get_wikitext2,
        "c4": get_c4,}
    if split not in ["train", "validation"]:
        raise ValueError(f"The split need to be 'train' or 'validation' but found {split}")
    if dataset_name not in get_dataset_map:
        raise ValueError(
            f"Expected a value in {list(get_dataset_map.keys())} but found {dataset_name}")
    get_dataset_fn = get_dataset_map[dataset_name]

    data = get_dataset_fn(
        tokenizer=tokenizer,
        nsamples=nsamples,
        seqlen=seqlen,
        split=split,
        fuse_sequences=fuse_sequences,
        seed=seed)

    # In case the dataset is loaded to be used with an fx.GraphModule, we need to add empty past_key_values inputs in the dataset.
    if require_fx:
        config = AutoConfig.from_pretrained(model_name_or_path)

        normalized_config_class = NormalizedConfigManager.get_normalized_config_class(
            config.model_type)
        normalized_config = normalized_config_class(config)

        num_heads = normalized_config.num_attention_heads
        if hasattr(normalized_config, "num_key_value_heads"):
            num_kv_heads = normalized_config.num_key_value_heads
        else:
            num_kv_heads = num_heads
        head_dim = normalized_config.hidden_size // num_heads
        num_layers = normalized_config.num_layers

        for sample in data:
            sample["past_key_values"] = tuple((
                torch.zeros(
                    1,
                    num_kv_heads,
                    0,
                    head_dim,
                    device=sample["input_ids"].device,
                    dtype=sample["input_ids"].dtype),
                torch.zeros(
                    1,
                    num_kv_heads,
                    0,
                    head_dim,
                    device=sample["input_ids"].device,
                    dtype=sample["input_ids"].dtype),
            ) for _ in range(num_layers))

    data = DatasetToDevice(data, device=device)

    return data



"""
Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
SPDX-License-Identifier: MIT
"""

import argparse
from datetime import datetime
from functools import partial
import json
import math
import os
import time

from dependencies import value
import diffusers
from diffusers import DiffusionPipeline
from diffusers import EulerDiscreteScheduler
from diffusers import StableDiffusionXLPipeline
from diffusers.models.attention_processor import Attention
import numpy as np
import packaging
import packaging.version
import pandas as pd
import torch
from torch import nn
from torchmetrics.image.fid import FrechetInceptionDistance
from tqdm import tqdm

from brevitas.core.stats.stats_op import NegativeMinOrZero
from brevitas.export.onnx.standard.qcdq.manager import StdQCDQONNXManager
from brevitas.graph.base import ModuleToModuleByClass
from brevitas.graph.calibrate import bias_correction_mode
from brevitas.graph.calibrate import calibration_mode
from brevitas.graph.calibrate import load_quant_model_mode
from brevitas.graph.equalize import activation_equalization_mode
from brevitas.graph.gptq import gptq_mode
from brevitas.graph.quantize import layerwise_quantize
from brevitas.inject.enum import StatsOp
from brevitas.nn.equalized_layer import EqualizedModule
from brevitas.utils.torch_utils import KwargsForwardHook
from brevitas_examples.common.generative.quantize import generate_quant_maps
from brevitas_examples.common.generative.quantize import generate_quantizers
from brevitas_examples.common.parse_utils import add_bool_arg
from brevitas_examples.common.parse_utils import quant_format_validator
from brevitas_examples.llm.llm_quant.export import BlockQuantProxyLevelManager
from brevitas_examples.stable_diffusion.mlperf_evaluation.accuracy import compute_mlperf_fid
from brevitas_examples.stable_diffusion.sd_quant.constants import SD_2_1_EMBEDDINGS_SHAPE
from brevitas_examples.stable_diffusion.sd_quant.constants import SD_XL_EMBEDDINGS_SHAPE
from brevitas_examples.stable_diffusion.sd_quant.export import export_onnx
from brevitas_examples.stable_diffusion.sd_quant.export import export_quant_params
from brevitas_examples.stable_diffusion.sd_quant.nn import AttnProcessor
from brevitas_examples.stable_diffusion.sd_quant.nn import AttnProcessor2_0
from brevitas_examples.stable_diffusion.sd_quant.nn import QuantAttention
from brevitas_examples.stable_diffusion.sd_quant.nn import QuantAttentionLast
from brevitas_examples.stable_diffusion.sd_quant.nn import QuantizableAttention
from brevitas_examples.stable_diffusion.sd_quant.utils import generate_latents
from brevitas_examples.stable_diffusion.sd_quant.utils import generate_unet_21_rand_inputs
from brevitas_examples.stable_diffusion.sd_quant.utils import generate_unet_xl_rand_inputs
from brevitas_examples.stable_diffusion.sd_quant.utils import unet_input_shape

diffusers_version = packaging.version.parse(diffusers.__version__)
TEST_SEED = 123456
torch.manual_seed(TEST_SEED)

NEGATIVE_PROMPTS = ["normal quality, low quality, worst quality, low res, blurry, nsfw, nude"]

CALIBRATION_PROMPTS = [
    'A man in a space suit playing a guitar, inspired by Cyril Rolando, highly detailed illustration, full color illustration, very detailed illustration, dan mumford and alex grey style',
    'a living room, bright modern Scandinavian style house, large windows, magazine photoshoot, 8k, studio lighting',
    'cute rabbit in a spacesuit',
    'minimalistic plolygon geometric car in brutalism warehouse, Rick Owens']

TESTING_PROMPTS = [
    'batman, cute modern disney style, Pixar 3d portrait, ultra detailed, gorgeous, 3d zbrush, trending on dribbble, 8k render',
    'A beautiful stack of rocks sitting on top of a beach, a picture, red black white golden colors, chakras, packshot, stock photo',
    'A painting of a fish on a black background, a digital painting, by Jason Benjamin, colorful vector illustration, mixed media style illustration, epic full color illustration, mascot illustration',
    'close up photo of a rabbit, forest in spring, haze, halation, bloom, dramatic atmosphere, centred, rule of thirds, 200mm 1.4f macro shot'
]


def load_calib_prompts(calib_data_path, sep="\t"):
    df = pd.read_csv(calib_data_path, sep=sep)
    lst = df["caption"].tolist()
    return lst


def run_test_inference(
        pipe,
        resolution,
        prompts,
        seeds,
        output_path,
        device,
        dtype,
        use_negative_prompts,
        guidance_scale,
        name_prefix=''):
    images = dict()
    with torch.no_grad():
        if not os.path.exists(output_path):
            os.mkdir(output_path)
        test_latents = generate_latents(seeds, device, dtype, unet_input_shape(resolution))
        neg_prompts = NEGATIVE_PROMPTS * len(seeds) if use_negative_prompts else []
        for prompt in prompts:
            prompt_images = pipe([prompt] * len(seeds),
                                 latents=test_latents,
                                 negative_prompt=neg_prompts,
                                 guidance_scale=guidance_scale).images
            images[prompt] = prompt_images

        i = 0
        for prompt, prompt_images in images.items():
            for image in prompt_images:
                file_path = os.path.join(output_path, f"{name_prefix}{i}.png")
                print(f"Saving to {file_path}")
                image.save(file_path)
                i += 1
    return images


def run_val_inference(
        pipe,
        resolution,
        prompts,
        seeds,
        device,
        dtype,
        use_negative_prompts,
        guidance_scale,
        total_steps,
        test_latents=None):
    with torch.no_grad():

        if test_latents is None:
            test_latents = generate_latents(seeds[0], device, dtype, unet_input_shape(resolution))

        neg_prompts = NEGATIVE_PROMPTS if use_negative_prompts else []
        for prompt in tqdm(prompts):
            # We don't want to generate any image, so we return only the latent encoding pre VAE
            pipe(
                prompt,
                negative_prompt=neg_prompts[0],
                latents=test_latents,
                output_type='latent',
                guidance_scale=guidance_scale,
                num_inference_steps=total_steps)


def main(args):

    dtype = getattr(torch, args.dtype)

    calibration_prompts = CALIBRATION_PROMPTS
    if args.calibration_prompt_path is not None:
        calibration_prompts = load_calib_prompts(args.calibration_prompt_path)

    assert args.calibration_prompt <= len(calibration_prompts) , f"Only {len(calibration_prompts)} prompts are available"
    calibration_prompts = calibration_prompts[:args.calibration_prompt]

    latents = None
    if args.path_to_latents is not None:
        latents = torch.load(args.path_to_latents).to(dtype)

    # Create output dir. Move to tmp if None
    ts = datetime.fromtimestamp(time.time())
    str_ts = ts.strftime("%Y%m%d_%H%M%S")
    output_dir = os.path.join(args.output_path, f'{str_ts}')
    os.mkdir(output_dir)
    print(f"Saving results in {output_dir}")

    # Dump args to json
    with open(os.path.join(output_dir, 'args.json'), 'w') as fp:
        json.dump(vars(args), fp)

    # Extend seeds based on batch_size
    test_seeds = [TEST_SEED] + [TEST_SEED + i for i in range(1, args.batch_size)]

    # Load model from float checkpoint
    print(f"Loading model from {args.model}...")
    variant = 'fp16' if dtype == torch.float16 else None
    pipe = DiffusionPipeline.from_pretrained(
        args.model, torch_dtype=dtype, variant=variant, use_safetensors=True)
    pipe.scheduler = EulerDiscreteScheduler.from_config(pipe.scheduler.config)
    pipe.vae.config.force_upcast = True
    is_mlperf_diffusers = diffusers_version == packaging.version.parse('0.21.2')

    AttClass = Attention
    if is_mlperf_diffusers:
        QuantAttClass = QuantAttention
        if args.share_qkv_quant:
            AttClass = QuantizableAttention
            rewriter = ModuleToModuleByClass(
                Attention,
                QuantizableAttention,
                query_dim=lambda module: module.to_q.in_features,
                dim_head=lambda module: math.ceil(1 / (module.scale ** 2)),
                bias=lambda module: hasattr(module.to_q, 'bias') and module.to_q.bias is not None,
                processor=AttnProcessor2_0(),
                dtype=dtype,
                norm_num_groups=lambda module: None
                if module.group_norm is None else module.group_norm.num_groups)
            rewriter.apply(pipe.unet)
    else:
        QuantAttClass = QuantAttentionLast
        if args.share_qkv_quant:
            pipe.fuse_qkv_projections()

    print(f"Model loaded from {args.model}.")

    # Move model to target device
    print(f"Moving model to {args.device}...")
    pipe = pipe.to(args.device)

    if args.prompt > 0 and not args.use_mlperf_inference:
        print(f"Running inference with prompt ...")
        testing_prompts = TESTING_PROMPTS[:args.prompt]
        float_images = run_test_inference(
            pipe,
            args.resolution,
            testing_prompts,
            test_seeds,
            output_dir,
            args.device,
            dtype,
            guidance_scale=args.guidance_scale,
            use_negative_prompts=args.use_negative_prompts,
            name_prefix='float_')

    # Detect Stable Diffusion XL pipeline
    is_sd_xl = isinstance(pipe, StableDiffusionXLPipeline)

    # Enable attention slicing
    if args.attention_slicing:
        pipe.enable_attention_slicing()

    # Extract list of layers to avoid
    blacklist = []
    non_blacklist = dict()
    for name, _ in pipe.unet.named_modules():
        if any(map(lambda x: x in name, args.quant_blacklist)):
            blacklist.append(name)
        else:
            if isinstance(_, (torch.nn.Linear, torch.nn.Conv2d)):
                name_to_add = name
                if name_to_add not in non_blacklist:
                    non_blacklist[name_to_add] = 1
                else:
                    non_blacklist[name_to_add] += 1
    print(f"Blacklisted layers: {set(blacklist)}")
    print(f"Non blacklisted layers: {set(non_blacklist.keys())}")

    # Make sure there all LoRA layers are fused first, otherwise raise an error
    for m in pipe.unet.modules():
        if hasattr(m, 'lora_layer') and m.lora_layer is not None:
            raise RuntimeError("LoRA layers should be fused in before calling into quantization.")

    if args.activation_equalization:
        pipe.set_progress_bar_config(disable=True)
        with torch.no_grad(), activation_equalization_mode(
                pipe.unet,
                alpha=args.act_eq_alpha,
                layerwise=True,
                blacklist_layers=blacklist if args.exclude_blacklist_act_eq else None,
                add_mul_node=True):
            # Workaround to expose `in_features` attribute from the Hook Wrapper
            for m in pipe.unet.modules():
                if isinstance(m, KwargsForwardHook) and hasattr(m.module, 'in_features'):
                    m.in_features = m.module.in_features
            total_steps = args.calibration_steps
            if args.dry_run or args.load_checkpoint is not None:
                calibration_prompts = [calibration_prompts[0]]
                total_steps = 1
            run_val_inference(
                pipe,
                args.resolution,
                calibration_prompts,
                test_seeds,
                args.device,
                dtype,
                total_steps=total_steps,
                use_negative_prompts=args.use_negative_prompts,
                test_latents=latents,
                guidance_scale=args.guidance_scale)

        # Workaround to expose `in_features` attribute from the EqualizedModule Wrapper
        for m in pipe.unet.modules():
            if isinstance(m, EqualizedModule) and hasattr(m.layer, 'in_features'):
                m.in_features = m.layer.in_features

    # Quantize model
    if args.quantize:

        @value
        def weight_bit_width(module):
            if isinstance(module, nn.Linear):
                return args.linear_weight_bit_width
            elif isinstance(module, nn.Conv2d):
                return args.conv_weight_bit_width
            else:
                raise RuntimeError(f"Module {module} not supported.")

        @value
        def input_bit_width(module):
            if isinstance(module, nn.Linear):
                return args.linear_input_bit_width
            elif isinstance(module, nn.Conv2d):
                return args.conv_input_bit_width
            else:
                raise RuntimeError(f"Module {module} not supported.")

        input_kwargs = dict()
        if args.input_scale_stats_op == 'minmax':

            @value
            def input_scale_stats_type():
                if args.input_quant_type == 'asym':
                    input_scaling_stats_op = StatsOp.MIN_MAX
                else:
                    input_scaling_stats_op = StatsOp.MAX
                return input_scaling_stats_op

            input_kwargs['scaling_stats_op'] = input_scale_stats_type

        if args.input_zp_stats_op == 'minmax':

            @value
            def input_zp_stats_type():
                if args.input_quant_type == 'asym':
                    zero_point_stats_impl = NegativeMinOrZero
                    return zero_point_stats_impl

            input_kwargs['zero_point_stats_impl'] = input_zp_stats_type

        sdpa_kwargs = dict()
        if args.sdpa_scale_stats_op == 'minmax':

            @value
            def sdpa_scale_stats_type():
                if args.sdpa_quant_type == 'asym':
                    sdpa_scaling_stats_op = StatsOp.MIN_MAX
                else:
                    sdpa_scaling_stats_op = StatsOp.MAX
                return sdpa_scaling_stats_op

            sdpa_kwargs['scaling_stats_op'] = sdpa_scale_stats_type

        if args.sdpa_zp_stats_op == 'minmax':

            @value
            def sdpa_zp_stats_type():
                if args.sdpa_quant_type == 'asym':
                    zero_point_stats_impl = NegativeMinOrZero
                    return zero_point_stats_impl

            sdpa_kwargs['zero_point_stats_impl'] = sdpa_zp_stats_type

        print("Applying model quantization...")
        quantizers = generate_quantizers(
            dtype=dtype,
            device=args.device,
            weight_bit_width=weight_bit_width,
            weight_quant_format=args.weight_quant_format,
            weight_quant_type=args.weight_quant_type,
            weight_param_method=args.weight_param_method,
            weight_scale_precision=args.weight_scale_precision,
            weight_quant_granularity=args.weight_quant_granularity,
            weight_group_size=args.weight_group_size,
            quantize_weight_zero_point=args.quantize_weight_zero_point,
            quantize_input_zero_point=args.quantize_input_zero_point,
            input_bit_width=input_bit_width,
            input_quant_format=args.input_quant_format,
            input_scale_type=args.input_scale_type,
            input_scale_precision=args.input_scale_precision,
            input_param_method=args.input_param_method,
            input_quant_type=args.input_quant_type,
            input_quant_granularity=args.input_quant_granularity,
            input_kwargs=input_kwargs)

        layer_map = generate_quant_maps(
            *quantizers, dtype, args.device, args.input_quant_format, False)

        linear_qkwargs = layer_map[torch.nn.Linear][1]
        linear_qkwargs[
            'input_quant'] = None if args.linear_input_bit_width == 0 else linear_qkwargs[
                'input_quant']
        linear_qkwargs[
            'weight_quant'] = None if args.linear_weight_bit_width == 0 else linear_qkwargs[
                'weight_quant']
        layer_map[torch.nn.Linear] = (layer_map[torch.nn.Linear][0], linear_qkwargs)

        conv_qkwargs = layer_map[torch.nn.Conv2d][1]
        conv_qkwargs[
            'input_quant'] = None if args.conv_input_bit_width == 0 else conv_qkwargs['input_quant']
        conv_qkwargs['weight_quant'] = None if args.conv_weight_bit_width == 0 else conv_qkwargs[
            'weight_quant']
        layer_map[torch.nn.Conv2d] = (layer_map[torch.nn.Conv2d][0], conv_qkwargs)

        if args.sdpa_bit_width > 0:
            # `args.weight_quant_granularity` must be compatible with `args.sdpa_quant_format`
            sdpa_quantizers = generate_quantizers(
                dtype=dtype,
                device=args.device,
                weight_bit_width=args.sdpa_bit_width,
                weight_quant_format=args.sdpa_quant_format,
                weight_quant_type=args.sdpa_quant_type,
                weight_param_method=args.weight_param_method,
                weight_scale_precision=args.weight_scale_precision,
                weight_quant_granularity=args.weight_quant_granularity,
                weight_group_size=args.weight_group_size,
                quantize_weight_zero_point=args.quantize_weight_zero_point,
                quantize_input_zero_point=args.quantize_sdpa_zero_point,
                input_bit_width=args.sdpa_bit_width,
                input_quant_format=args.sdpa_quant_format,
                input_scale_type=args.sdpa_scale_type,
                input_scale_precision=args.sdpa_scale_precision,
                input_param_method=args.sdpa_param_method,
                input_quant_type=args.sdpa_quant_type,
                input_quant_granularity=args.sdpa_quant_granularity,
                input_kwargs=sdpa_kwargs)
            # We generate all quantizers, but we are only interested in activation quantization for
            # the output of softmax and the output of QKV
            input_quant = sdpa_quantizers[0]
            if is_mlperf_diffusers:
                extra_kwargs = {}
                query_lambda = lambda module: module.to_qkv.in_features if hasattr(
                    module, 'to_qkv') else module.to_q.in_features
            else:
                extra_kwargs = {
                    'fuse_qkv':
                        args.share_qkv_quant,
                    'cross_attention_dim':
                        lambda module: module.cross_attention_dim
                        if module.is_cross_attention else None}
                query_lambda = lambda module: module.query_dim
            rewriter = ModuleToModuleByClass(
                AttClass,
                QuantAttClass,
                matmul_input_quant=input_quant,
                query_dim=query_lambda,
                dim_head=lambda module: math.ceil(1 / (module.scale ** 2)),
                processor=AttnProcessor(),
                is_equalized=args.activation_equalization,
                **extra_kwargs)
            import brevitas.config as config
            config.IGNORE_MISSING_KEYS = True
            pipe.unet = rewriter.apply(pipe.unet)
            config.IGNORE_MISSING_KEYS = False
            pipe.unet = pipe.unet.to(args.device)
            pipe.unet = pipe.unet.to(dtype)

            if args.override_conv_quant_config:
                print(
                    f"Overriding Conv2d quantization to weights: {sdpa_quantizers[1]}, inputs: {sdpa_quantizers[2]}"
                )
                conv_qkwargs = layer_map[torch.nn.Conv2d][1]
                conv_qkwargs['input_quant'] = sdpa_quantizers[2]
                conv_qkwargs['weight_quant'] = sdpa_quantizers[1]
                layer_map[torch.nn.Conv2d] = (layer_map[torch.nn.Conv2d][0], conv_qkwargs)

        pipe.unet = layerwise_quantize(
            model=pipe.unet, compute_layer_map=layer_map, name_blacklist=blacklist)
        print("Model quantization applied.")

        pipe.set_progress_bar_config(disable=True)

        with torch.no_grad():
            run_val_inference(
                pipe,
                args.resolution, [calibration_prompts[0]],
                test_seeds,
                args.device,
                dtype,
                total_steps=1,
                use_negative_prompts=args.use_negative_prompts,
                test_latents=latents,
                guidance_scale=args.guidance_scale)

        if args.load_checkpoint is not None:
            with load_quant_model_mode(pipe.unet):
                pipe = pipe.to('cpu')
                print(f"Loading checkpoint: {args.load_checkpoint}... ", end="")
                pipe.unet.load_state_dict(torch.load(args.load_checkpoint, map_location='cpu'))
                print(f"Checkpoint loaded!")
            pipe = pipe.to(args.device)
        elif not args.dry_run:
            # Model needs calibration if any of its activation quantizers are 'static'
            activation_bw = [
                args.linear_input_bit_width,
                args.conv_input_bit_width,
                args.sdpa_bit_width,]
            activation_st = [
                args.input_scale_type,
                args.input_scale_type,
                args.sdpa_scale_type,]
            needs_calibration = any(
                map(lambda b, st: (b > 0) and st == 'static', activation_bw, activation_st))
            if needs_calibration:
                print("Applying activation calibration")
                with torch.no_grad(), calibration_mode(pipe.unet):
                    run_val_inference(
                        pipe,
                        args.resolution,
                        calibration_prompts,
                        test_seeds,
                        args.device,
                        dtype,
                        total_steps=args.calibration_steps,
                        use_negative_prompts=args.use_negative_prompts,
                        test_latents=latents,
                        guidance_scale=args.guidance_scale)

            if args.gptq:
                print("Applying GPTQ. It can take several hours")
                with torch.no_grad(), gptq_mode(pipe.unet,
                            create_weight_orig=False,
                            use_quant_activations=False,
                            return_forward_output=True,
                            act_order=True) as gptq:
                    for _ in tqdm(range(gptq.num_layers)):
                        run_val_inference(
                            pipe,
                            args.resolution,
                            calibration_prompts,
                            test_seeds,
                            args.device,
                            dtype,
                            total_steps=args.calibration_steps,
                            use_negative_prompts=args.use_negative_prompts,
                            test_latents=latents,
                            guidance_scale=args.guidance_scale)
                        gptq.update()
                        torch.cuda.empty_cache()
            if args.bias_correction:
                print("Applying bias correction")
                with torch.no_grad(), bias_correction_mode(pipe.unet):
                    run_val_inference(
                        pipe,
                        args.resolution,
                        calibration_prompts,
                        test_seeds,
                        args.device,
                        dtype,
                        total_steps=args.calibration_steps,
                        use_negative_prompts=args.use_negative_prompts,
                        test_latents=latents,
                        guidance_scale=args.guidance_scale)

    if args.vae_fp16_fix and is_sd_xl:
        vae_fix_scale = 128
        layer_whitelist = [
            "decoder.up_blocks.2.upsamplers.0.conv",
            "decoder.up_blocks.3.resnets.0.conv2",
            "decoder.up_blocks.3.resnets.1.conv2",
            "decoder.up_blocks.3.resnets.2.conv2"]
        #layer_whitelist = [
        #    "decoder.up_blocks.3.resnets.0.conv_shortcut",
        #    "decoder.up_blocks.3.resnets.0.conv2",
        #    "decoder.up_blocks.3.resnets.1.conv2",
        #    "decoder.up_blocks.3.resnets.2.conv2"]
        corrected_layers = []
        with torch.no_grad():
            for name, module in pipe.vae.named_modules():
                if name in layer_whitelist:
                    corrected_layers.append(name)
                    module.weight /= vae_fix_scale
                    if module.bias is not None:
                        module.bias /= vae_fix_scale
        print(f"Corrected layers in VAE: {corrected_layers}")

    if args.checkpoint_name is not None and args.load_checkpoint is None:
        torch.save(pipe.unet.state_dict(), os.path.join(output_dir, args.checkpoint_name))
        if args.vae_fp16_fix:
            torch.save(
                pipe.vae.state_dict(), os.path.join(output_dir, f"vae_{args.checkpoint_name}"))

    if args.export_target:
        # Move to cpu and to float32 to enable CPU export
        if args.export_cpu_float32:
            pipe.unet.to('cpu').to(torch.float32)
        pipe.unet.eval()
        device = next(iter(pipe.unet.parameters())).device
        dtype = next(iter(pipe.unet.parameters())).dtype

        # Define tracing input
        if is_sd_xl:
            generate_fn = generate_unet_xl_rand_inputs
            shape = SD_XL_EMBEDDINGS_SHAPE
        else:
            generate_fn = generate_unet_21_rand_inputs
            shape = SD_2_1_EMBEDDINGS_SHAPE
        trace_inputs = generate_fn(
            embedding_shape=shape,
            unet_input_shape=unet_input_shape(args.resolution),
            device=device,
            dtype=dtype)

        if args.export_target == 'onnx':
            if args.weight_quant_granularity == 'per_group':
                export_manager = BlockQuantProxyLevelManager
            else:
                export_manager = StdQCDQONNXManager
                export_manager.change_weight_export(export_weight_q_node=args.export_weight_q_node)
            export_onnx(pipe, trace_inputs, output_dir, export_manager)
        if args.export_target == 'params_only':
            pipe.to('cpu')
            export_quant_params(pipe, output_dir, export_vae=args.vae_fp16_fix)

    # Perform inference
    if args.prompt > 0 and not args.dry_run:
        # with brevitas_proxy_inference_mode(pipe.unet):
        if args.use_mlperf_inference:
            print(f"Computing accuracy with MLPerf pipeline")
            compute_mlperf_fid(
                args.model,
                args.path_to_coco,
                pipe,
                args.prompt,
                output_dir,
                args.device,
                not args.vae_fp16_fix)
        else:
            print(f"Computing accuracy on default prompt")
            testing_prompts = TESTING_PROMPTS[:args.prompt]
            assert args.prompt <= len(TESTING_PROMPTS), f"Only {len(TESTING_PROMPTS)} prompts are available"

            quant_images = run_test_inference(
                pipe,
                args.resolution,
                testing_prompts,
                test_seeds,
                output_dir,
                args.device,
                dtype,
                use_negative_prompts=args.use_negative_prompts,
                guidance_scale=args.guidance_scale,
                name_prefix='quant_')

            float_images_values = float_images.values()
            float_images_values = [x for x_nested in float_images_values for x in x_nested]
            float_images_values = torch.tensor([np.array(image) for image in float_images_values])
            float_images_values = float_images_values.permute(0, 3, 1, 2)

            quant_images_values = quant_images.values()
            quant_images_values = [x for x_nested in quant_images_values for x in x_nested]
            quant_images_values = torch.tensor([np.array(image) for image in quant_images_values])
            quant_images_values = quant_images_values.permute(0, 3, 1, 2)

            fid = FrechetInceptionDistance(normalize=False)
            fid.update(float_images_values, real=True)
            fid.update(quant_images_values, real=False)
            print(f"FID: {float(fid.compute())}")


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='Stable Diffusion quantization')
    parser.add_argument(
        '-m',
        '--model',
        type=str,
        default='/scratch/hf_models/stable-diffusion-2-1-base',
        help='Path or name of the model.')
    parser.add_argument(
        '-d', '--device', type=str, default='cuda:0', help='Target device for quantized model.')
    parser.add_argument(
        '-b',
        '--batch-size',
        type=int,
        default=2,
        help='How many seeds to use for each image during validation. Default: 2')
    parser.add_argument(
        '--prompt', type=int, default=4, help='Number of prompt to use for testing. Default: 4')
    parser.add_argument(
        '--calibration-prompt',
        type=int,
        default=2,
        help='Number of prompt to use for calibration. Default: 2')
    parser.add_argument(
        '--calibration-prompt-path', type=str, default=None, help='Path to calibration prompt')
    parser.add_argument(
        '--checkpoint-name',
        type=str,
        default=None,
        help=
        'Name to use to store the checkpoint in the output dir. If not provided, no checkpoint is saved.'
    )
    parser.add_argument(
        '--load-checkpoint',
        type=str,
        default=None,
        help='Path to checkpoint to load. If provided, PTQ techniques are skipped.')
    parser.add_argument(
        '--path-to-latents',
        type=str,
        default=None,
        help=
        'Load pre-defined latents. If not provided, they are generated based on an internal seed.')
    parser.add_argument(
        '--path-to-coco',
        type=str,
        default=None,
        help=
        'Path to MLPerf compliant Coco dataset. Used when the --use-mlperf flag is set. Default: None'
    )
    parser.add_argument(
        '--resolution',
        type=int,
        default=512,
        help='Resolution along height and width dimension. Default: 512.')
    parser.add_argument('--guidance-scale', type=float, default=7.5, help='Guidance scale.')
    parser.add_argument(
        '--calibration-steps', type=int, default=8, help='Steps used during calibration')
    add_bool_arg(
        parser,
        'output-path',
        str_true=True,
        default='.',
        help='Path where to generate output folder.')
    add_bool_arg(parser, 'quantize', default=True, help='Toggle quantization. Default: Enabled')
    add_bool_arg(
        parser,
        'activation-equalization',
        default=False,
        help='Toggle Activation Equalization. Default: Disabled')
    add_bool_arg(parser, 'gptq', default=False, help='Toggle gptq. Default: Disabled')
    add_bool_arg(
        parser, 'bias-correction', default=True, help='Toggle bias-correction. Default: Enabled')
    parser.add_argument(
        '--dtype',
        default='float16',
        choices=['float32', 'float16', 'bfloat16'],
        help='Model Dtype, choices are float32, float16, bfloat16. Default: float16')
    add_bool_arg(
        parser,
        'attention-slicing',
        default=False,
        help='Enable attention slicing. Default: Disabled')
    parser.add_argument(
        '--export-target',
        type=str,
        default='',
        choices=['', 'onnx', 'params_only'],
        help='Target export flow.')
    add_bool_arg(
        parser,
        'export-weight-q-node',
        default=False,
        help=
        'Enable export of floating point weights + QDQ rather than integer weights + DQ. Default: Disabled'
    )
    parser.add_argument(
        '--conv-weight-bit-width', type=int, default=8, help='Weight bit width. Default: 8.')
    parser.add_argument(
        '--linear-weight-bit-width', type=int, default=8, help='Weight bit width. Default: 8.')
    parser.add_argument(
        '--conv-input-bit-width',
        type=int,
        default=0,
        help='Input bit width. Default: 0 (not quantized)')
    parser.add_argument(
        '--act-eq-alpha',
        type=float,
        default=0.9,
        help='Alpha for activation equalization. Default: 0.9')
    parser.add_argument(
        '--linear-input-bit-width',
        type=int,
        default=0,
        help='Input bit width. Default: 0 (not quantized).')
    parser.add_argument(
        '--weight-param-method',
        type=str,
        default='stats',
        choices=['stats', 'mse'],
        help='How scales/zero-point are determined. Default: stats.')
    parser.add_argument(
        '--input-param-method',
        type=str,
        default='stats',
        choices=['stats', 'mse'],
        help='How scales/zero-point are determined. Default: stats.')
    parser.add_argument(
        '--input-scale-stats-op',
        type=str,
        default='minmax',
        choices=['minmax', 'percentile'],
        help='Define what statics op to use for input scale. Default: minmax.')
    parser.add_argument(
        '--input-zp-stats-op',
        type=str,
        default='minmax',
        choices=['minmax', 'percentile'],
        help='Define what statics op to use for input zero point. Default: minmax.')
    parser.add_argument(
        '--weight-scale-precision',
        type=str,
        default='float_scale',
        choices=['float_scale', 'po2_scale'],
        help='Whether scale is a float value or a po2. Default: float_scale.')
    parser.add_argument(
        '--input-scale-precision',
        type=str,
        default='float_scale',
        choices=['float_scale', 'po2_scale'],
        help='Whether scale is a float value or a po2. Default: float_scale.')
    parser.add_argument(
        '--weight-quant-type',
        type=str,
        default='asym',
        choices=['sym', 'asym'],
        help='Weight quantization type. Default: asym.')
    parser.add_argument(
        '--input-quant-type',
        type=str,
        default='asym',
        choices=['sym', 'asym'],
        help='Input quantization type. Default: asym.')
    parser.add_argument(
        '--weight-quant-format',
        type=quant_format_validator,
        default='int',
        help=
        'Weight quantization type. Either int or eXmY, with X+Y==weight_bit_width-1. It\'s possible to add float_ocp_ or float_fnuz_ before the exponent/mantissa bitwidth. Default: int.'
    )
    parser.add_argument(
        '--input-quant-format',
        type=quant_format_validator,
        default='int',
        help=
        'Input quantization type. Either int or eXmY, with X+Y==input_bit_width-1. It\'s possible to add float_ocp_ or float_fnuz_ before the exponent/mantissa bitwidth. Default: int.'
    )
    parser.add_argument(
        '--weight-quant-granularity',
        type=str,
        default='per_channel',
        choices=['per_channel', 'per_tensor', 'per_group'],
        help='Granularity for scales/zero-point of weights. Default: per_channel.')
    parser.add_argument(
        '--input-quant-granularity',
        type=str,
        default='per_tensor',
        choices=['per_tensor'],
        help='Granularity for scales/zero-point of inputs. Default: per_tensor.')
    parser.add_argument(
        '--input-scale-type',
        type=str,
        default='static',
        choices=['static', 'dynamic'],
        help='Whether to do static or dynamic input quantization. Default: static.')
    parser.add_argument(
        '--weight-group-size',
        type=int,
        default=16,
        help='Group size for per_group weight quantization. Default: 16.')
    parser.add_argument(
        '--sdpa-bit-width',
        type=int,
        default=0,
        help='Scaled dot product attention bit width. Default: 0 (not quantized).')
    parser.add_argument(
        '--sdpa-param-method',
        type=str,
        default='stats',
        choices=['stats', 'mse'],
        help=
        'How scales/zero-point are determined for scaled dot product attention. Default: %(default)s.'
    )
    parser.add_argument(
        '--sdpa-scale-stats-op',
        type=str,
        default='minmax',
        choices=['minmax', 'percentile'],
        help=
        'Define what statistics op to use for scaled dot product attention scale. Default: %(default)s.'
    )
    parser.add_argument(
        '--sdpa-zp-stats-op',
        type=str,
        default='minmax',
        choices=['minmax', 'percentile'],
        help=
        'Define what statistics op to use for scaled dot product attention zero point. Default: %(default)s.'
    )
    parser.add_argument(
        '--sdpa-scale-precision',
        type=str,
        default='float_scale',
        choices=['float_scale', 'po2_scale'],
        help=
        'Whether the scaled dot product attention scale is a float value or a po2. Default: %(default)s.'
    )
    parser.add_argument(
        '--sdpa-quant-type',
        type=str,
        default='sym',
        choices=['sym', 'asym'],
        help='Scaled dot product attention quantization type. Default: %(default)s.')
    parser.add_argument(
        '--sdpa-quant-format',
        type=quant_format_validator,
        default='int',
        help=
        'Scaled dot product attention quantization format. Either int or eXmY, with X+Y==input_bit_width-1. It\'s possible to add float_ocp_ or float_fnuz_ before the exponent/mantissa bitwidth. Default: %(default)s.'
    )
    parser.add_argument(
        '--sdpa-quant-granularity',
        type=str,
        default='per_tensor',
        choices=['per_tensor'],
        help=
        'Granularity for scales/zero-point of scaled dot product attention. Default: %(default)s.')
    parser.add_argument(
        '--sdpa-scale-type',
        type=str,
        default='static',
        choices=['static', 'dynamic'],
        help=
        'Whether to do static or dynamic scaled dot product attention quantization. Default: %(default)s.'
    )
    parser.add_argument(
        '--quant-blacklist',
        type=str,
        default=['time_emb'],
        nargs='*',
        metavar='NAME',
        help='A list of module names to exclude from quantization. Default: %(default)s')
    add_bool_arg(
        parser,
        'quantize-weight-zero-point',
        default=True,
        help='Quantize weight zero-point. Default: Enabled')
    add_bool_arg(
        parser,
        'exclude-blacklist-act-eq',
        default=False,
        help='Exclude unquantized layers from activation equalization. Default: Disabled')
    add_bool_arg(
        parser,
        'quantize-input-zero-point',
        default=False,
        help='Quantize input zero-point. Default: Enabled')
    add_bool_arg(
        parser,
        'quantize-sdpa-zero-point',
        default=False,
        help='Quantize scaled dot product attention zero-point. Default: %(default)s')
    add_bool_arg(
        parser, 'export-cpu-float32', default=False, help='Export FP32 on CPU. Default: Disabled')
    add_bool_arg(
        parser,
        'use-mlperf-inference',
        default=False,
        help='Evaluate FID score with MLPerf pipeline. Default: False')
    add_bool_arg(
        parser,
        'use-negative-prompts',
        default=True,
        help='Use negative prompts during generation/calibration. Default: Enabled')
    add_bool_arg(
        parser,
        'dry-run',
        default=False,
        help='Generate a quantized model without any calibration. Default: Disabled')
    add_bool_arg(
        parser,
        'override-conv-quant-config',
        default=False,
        help='Quantize Convolutions in the same way as SDP (i.e., FP8). Default: Disabled')
    add_bool_arg(
        parser,
        'vae-fp16-fix',
        default=False,
        help='Rescale the VAE to not go NaN with FP16. Default: Disabled')
    add_bool_arg(
        parser,
        'share-qkv-quant',
        default=False,
        help='Share QKV/KV quantization. Default: Disabled')
    args = parser.parse_args()
    print("Args: " + str(vars(args)))
    main(args)



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

from time import sleep
import warnings

import torch
import torchvision

from brevitas_examples.imagenet_classification import models

TIME_BETWEEN_RETRIES = 2 * 60  # in seconds


def get_model_config(model_name):
    config = dict()
    # Set-up config parameters
    if model_name == 'inception_v3' or model_name == 'googlenet':
        config['inception_preprocessing'] = True
    else:
        config['inception_preprocessing'] = False

    if model_name == 'inception_v3':
        input_shape = 299
        resize_shape = 342
    else:
        input_shape = 224
        resize_shape = 256
    config.update({'resize_shape': resize_shape, 'center_crop_shape': input_shape})
    return config


def get_torchvision_model(model_name):
    model_fn = getattr(torchvision.models, model_name)
    if model_name == 'inception_v3' or model_name == 'googlenet':
        model = model_fn(pretrained=True, transform_input=False)
    else:
        model = model_fn(pretrained=True)

    return model


def get_quant_model(model_name, bit_width):
    model_fn = getattr(models, model_name)
    model = model_fn(float_pretrained=True, bit_width=bit_width)
    return model


def add_bool_arg(parser, name, default, help, str_true=False):
    dest = name.replace('-', '_')
    group = parser.add_mutually_exclusive_group(required=False)
    if str_true:
        group.add_argument('--' + name, dest=dest, type=str, help=help)
    else:
        group.add_argument('--' + name, dest=dest, action='store_true', help='Enable ' + help)
    group.add_argument('--no-' + name, dest=dest, action='store_false', help='Disable ' + help)
    parser.set_defaults(**{dest: default})


def get_gpu_index(idx):
    gpu_world_size = torch.cuda.device_count()
    if gpu_world_size == 0:
        # If no GPU is available, execute on CPU
        return None

    gpu_id = idx % gpu_world_size
    try:
        # If GPUtil is instealled, fetch the next available GPU
        return get_next_available_gpu(gpu_id)
    except ModuleNotFoundError:
        warnings.warn("GPUtil not installed, multiple jobs might end up on the same GPU")
        # If GPUtil is not installed, cycle through GPUs
        return gpu_id


def get_next_available_gpu(gpu_id):
    import GPUtil
    GPUs = GPUtil.getGPUs()
    available_gpus = GPUtil.getAvailability(
        GPUs, maxLoad=0.5, maxMemory=0.5, includeNan=False, excludeID=[], excludeUUID=[])
    if available_gpus[gpu_id] == 0:
        next_available_gpu = GPUtil.getAvailable(
            order='memory',
            limit=3,
            maxLoad=0.5,
            maxMemory=0.5,
            includeNan=False,
            excludeID=[],
            excludeUUID=[])
        while len(next_available_gpu) == 0:
            sleep(TIME_BETWEEN_RETRIES)
            next_available_gpu = GPUtil.getAvailable(
                order='memory',
                limit=3,
                maxLoad=0.5,
                maxMemory=0.5,
                includeNan=False,
                excludeID=[],
                excludeUUID=[])

        gpu_id = next_available_gpu[0]
    return gpu_id



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

import torch
from torch.nn import Parameter

from brevitas.utils.torch_utils import compute_channel_view_shape


def mul_add_from_bn(bn_mean, bn_var, bn_eps, bn_weight, bn_bias):
    denom = torch.sqrt(bn_var + bn_eps)
    mul_factor = bn_weight / denom
    add_factor = -bn_mean * mul_factor + bn_bias
    return mul_factor, add_factor


def merge_bn(layer, bn, output_channel_dim=0):
    from brevitas.proxy.parameter_quant import BiasQuantProxyFromInjectorBase
    from brevitas.proxy.parameter_quant import WeightQuantProxyFromInjectorBase
    out = mul_add_from_bn(
        bn_mean=bn.running_mean,
        bn_var=bn.running_var,
        bn_eps=bn.eps,
        bn_weight=bn.weight.data.clone(),
        bn_bias=bn.bias.data.clone())
    mul_factor, add_factor = out
    out_ch_weight_shape = compute_channel_view_shape(layer.weight, output_channel_dim)
    layer.weight.data.mul_(mul_factor.view(out_ch_weight_shape))
    if layer.bias is not None:
        out_ch_bias_shape = compute_channel_view_shape(layer.bias, channel_dim=0)
        layer.bias.data.mul_(mul_factor.view(out_ch_bias_shape))
        layer.bias.data.add_(add_factor.view(out_ch_bias_shape))
    else:
        layer.bias = Parameter(add_factor)
    if (hasattr(layer, 'weight_quant') and
            isinstance(layer.weight_quant, WeightQuantProxyFromInjectorBase)):
        layer.weight_quant.init_tensor_quant()
    if (hasattr(layer, 'bias_quant') and
            isinstance(layer.bias_quant, BiasQuantProxyFromInjectorBase)):
        layer.bias_quant.init_tensor_quant()


def rename_state_dict_by_prefix(old_prefix, new_prefix, state_dict):
    keys_map = {}
    for k in state_dict.keys():
        if k.startswith(old_prefix):
            new_key = new_prefix + k[len(old_prefix):]
            keys_map[k] = new_key
    for old_key in keys_map.keys():
        state_dict[keys_map[old_key]] = state_dict.pop(old_key)


def rename_state_dict_by_postfix(old_postfix, new_postfix, state_dict):
    keys_map = {}
    for k in state_dict.keys():
        if k.endswith(old_postfix):
            new_key = k[:len(k) - len(old_postfix)] + new_postfix
            keys_map[k] = new_key
    for old_key in keys_map.keys():
        state_dict[keys_map[old_key]] = state_dict.pop(old_key)


def check_tensors_same_ptr(tensor_list):
    pointers = []
    for t in tensor_list:
        if hasattr(t, 'data_ptr'):
            ptr = t.data_ptr()
            pointers.append(ptr)
        elif hasattr(t, 'value') and hasattr(t.value, 'data_ptr'):
            pointers.append(t.value.data_ptr())
        else:
            return False
    return all(p == pointers[0] for p in pointers)



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

from brevitas.core.restrict_val import RestrictValueType
from brevitas.quant import Int8ActPerTensorFloat
from brevitas.quant import Int8WeightPerTensorFloat
from brevitas.quant import Uint8ActPerTensorFloat


class CommonIntWeightPerTensorQuant(Int8WeightPerTensorFloat):
    """
    Common per-tensor weight quantizer with bit-width set to None so that it's forced to be
    specified by each layer.
    """
    scaling_min_val = 2e-16
    bit_width = None


class CommonIntWeightPerChannelQuant(CommonIntWeightPerTensorQuant):
    """
    Common per-channel weight quantizer with bit-width set to None so that it's forced to be
    specified by each layer.
    """
    scaling_per_output_channel = True


class CommonIntActQuant(Int8ActPerTensorFloat):
    """
    Common signed act quantizer with bit-width set to None so that it's forced to be specified by
    each layer.
    """
    scaling_min_val = 2e-16
    bit_width = None
    restrict_scaling_type = RestrictValueType.LOG_FP


class CommonUintActQuant(Uint8ActPerTensorFloat):
    """
    Common unsigned act quantizer with bit-width set to None so that it's forced to be specified by
    each layer.
    """
    scaling_min_val = 2e-16
    bit_width = None
    restrict_scaling_type = RestrictValueType.LOG_FP



import torch.nn as nn
import torch

def quantize(tensor, scale, zero_point, is_asym=False):
    if is_asym:
        clamp_min, clamp_max = torch.tensor(0.), torch.tensor(15)
    else:
        clamp_min, clamp_max = torch.tensor(-7), torch.tensor(7)
    quant_tensor = torch.clamp(torch.round(tensor/scale + zero_point), clamp_min, clamp_max) 
    return quant_tensor

def dequantize(tensor, scale, zero_point):
    return (tensor - zero_point) * scale


class QuantLinear(nn.Module):
    def __init__(self, in_ch, out_ch, quant_param):
        super().__init__()
        self.out_ch = out_ch
        self.in_ch = in_ch
        self.linear = nn.Linear(in_ch, out_ch)
        self.group_dim = 1
        self.group_size = 32 # Can be parametrized
        
        # Fields name are temporary
        # weight_scale has shape [out_ch, in_ch//self.group_size, 1]
        po2_weight_scale = torch.tensor(quant_param['weight_scale']).view(quant_param['weight_scale_shape'])
        assert po2_weight_scale.dtype == torch.int8
        assert po2_weight_scale.shape == [out_ch, in_ch//self.group_size, 1]
        weight_scale = torch.pow(2, po2_weight_scale.to(torch.float16)) # Assuming fp16 dtype

        # weight_zp has shape [out_ch, in_ch//self.group_size, 1]
        weight_zp = torch.tensor(quant_param['weight_zp']).view(quant_param['weight_zp_shape'])
        assert quant_param['weight_zp_dtype'] == 'torch.int8', f"Weight Zero-Point dtype should be 'torch.int8', found: {quant_param['weight_zp_dype']}"
        assert weight_zp.shape == [out_ch, in_ch//self.group_size, 1]
        assert torch.max(weight_zp) <= 15., "Max value is above uint4"

        self.register_buffer('weight_scale', weight_scale)
        self.register_buffer('weight_zp', weight_zp)


    # I.e., "fake quantization"
    def qdq_forward(self, x):
        weight = weight.view(self.out_ch, self.in_ch//self.group_size, self.group_size)
        quant_weight = quantize(weight, self.weight_scale, self.weight_zp, is_asym=True)
        dequantized_weight = dequantize(quant_weight, self.weight_scale, self.weight_zp)
        # Go from  [out_ch, in_ch // group_size, group_size] to [out_ch, in_ch]
        dequantized_weight = dequantized_weight.flatten(1)
        out = torch.nn.functional.linear(x, dequantized_weight, self.linear.bias)
        return out



"""
Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
"""

from torch import nn

from brevitas.core.function_wrapper.shape import OverOutputFeaturesView
from brevitas.core.function_wrapper.shape import OverTensorView
from brevitas.core.scaling.runtime import RuntimeDynamicGroupStatsScaling
from brevitas.core.stats import AbsMinMax
from brevitas.core.stats import NegativeMinOrZero
from brevitas.core.stats.stats_op import HalfQuadraticOptimizerZeroPoint
from brevitas.core.stats.stats_wrapper import SCALAR_SHAPE
from brevitas.core.zero_point import StatsFromParameterZeroPoint
from brevitas.inject import ExtendedInjector
from brevitas.inject import this
from brevitas.inject import value
from brevitas.inject.enum import ScalingPerOutputType
from brevitas.proxy.groupwise_float_parameter_quant import \
    GroupwiseWeightFloatQuantProxyFromInjector
from brevitas.proxy.groupwise_float_runtime_quant import GroupwiseActFloatQuantProxyFromInjector
from brevitas.proxy.groupwise_int_parameter_quant import GroupwiseWeightQuantProxyFromInjector
from brevitas.proxy.groupwise_int_runtime_quant import GroupwiseActQuantProxyFromInjector
from brevitas.proxy.runtime_quant import DynamicActQuantProxyFromInjector
from brevitas.quant.base import HQOWeightZeroPoint
from brevitas.quant.experimental.float import Fp8e4m3ActPerTensorFloat
from brevitas.quant.experimental.float import Fp8e4m3WeightPerChannelFloat
from brevitas.quant.experimental.float_quant_ocp import Fp8e4m3OCPActPerTensorFloat
from brevitas.quant.scaled_int import Int8ActPerTensorFloat
from brevitas.quant.scaled_int import Int8WeightPerChannelFloat
from brevitas.quant.scaled_int import Int8WeightPerChannelFloatHQO
from brevitas.quant.shifted_scaled_int import ShiftedUint8ActPerTensorFloat
from brevitas.quant.shifted_scaled_int import ShiftedUint8WeightPerChannelFloat

from .quant_blocks import *


class DynamicActProxyMixin(ExtendedInjector):
    proxy_class = DynamicActQuantProxyFromInjector


class IntWeightSymmetricGroupQuant(Int8WeightPerChannelFloat):
    """
    Block / group / vector signed symmetric int weight quantizer with float scales.
    We inherit from a per-channel quantizer to re-use some underlying machinery.
    """
    proxy_class = GroupwiseWeightQuantProxyFromInjector
    scaling_per_output_type = ScalingPerOutputType.GROUP


class Fp8e4m3WeightSymmetricGroupQuant(Fp8e4m3WeightPerChannelFloat):
    """
    Block / group / vector signed symmetric e4m3 weight quantizer with float scales.
    We inherit from a per-channel quantizer to re-use some underlying machinery.
    """
    proxy_class = GroupwiseWeightFloatQuantProxyFromInjector
    scaling_per_output_type = ScalingPerOutputType.GROUP


class Int8DynamicActPerTensorFloat(DynamicActProxyMixin, Int8ActPerTensorFloat):
    """
    Symmetric quantizer with per tensor dynamic scale.
    """
    scaling_impl = RuntimeDynamicStatsScaling
    scaling_stats_input_view_shape_impl = OverTensorView
    scaling_stats_op = 'min_max'
    dynamic_scaling_broadcastable_fn = lambda x, shape: x.view(SCALAR_SHAPE)


class Int8DynamicActPerRowFloat(DynamicActProxyMixin, Int8ActPerTensorFloat):
    """
    Symmetric quantizer with per row dynamic scale.
    """
    scaling_impl = RuntimeDynamicStatsScaling
    scaling_stats_input_view_shape_impl = OverOutputFeaturesView
    scaling_stats_op = 'min_max'
    scaling_per_output_channel = True


class Int8DynamicActPerGroupFloat(DynamicActProxyMixin, Int8ActPerTensorFloat):
    """
    Symmetric quantizer with per group scale.
    """
    proxy_class = GroupwiseActQuantProxyFromInjector
    scaling_impl = RuntimeDynamicGroupStatsScaling
    scaling_stats_op = 'min_max'
    scaling_per_output_type = ScalingPerOutputType.GROUP


class ShiftedUint8DynamicActPerTensorFloat(DynamicActProxyMixin, ShiftedUint8ActPerTensorFloat):
    """
    Symmetric quantizer with per tensor dynamic scale.
    """
    scaling_impl = RuntimeDynamicStatsScaling
    scaling_stats_input_view_shape_impl = OverTensorView
    scaling_stats_op = 'min_max'
    zero_point_impl = RuntimeDynamicStatsZeroPoint
    zero_point_stats_impl = NegativeMinOrZero
    dynamic_scaling_broadcastable_fn = lambda x, shape: x.view(SCALAR_SHAPE)


class ShiftedUint8DynamicActPerRowFloat(DynamicActProxyMixin, ShiftedUint8ActPerTensorFloat):
    """
    Asymmetric quantizer with per row dynamic scale.
    """
    scaling_impl = RuntimeDynamicStatsScaling
    scaling_stats_input_view_shape_impl = OverOutputFeaturesView
    scaling_stats_op = 'min_max'
    scaling_per_output_channel = True
    zero_point_impl = RuntimeDynamicStatsZeroPoint
    zero_point_stats_impl = NegativeMinOrZero


class Fp8e4m3DynamicActPerGroupFloat(DynamicActProxyMixin, Fp8e4m3ActPerTensorFloat):
    """
    Symmetric quantizer with per group scale.
    """
    proxy_class = GroupwiseActFloatQuantProxyFromInjector
    scaling_impl = RuntimeDynamicGroupStatsScaling
    scaling_per_output_type = ScalingPerOutputType.GROUP
    scaling_stats_op = 'min_max'
